/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.10 $";
#endif

/*
 * Run a Command
 */

#include "defs.h"
#include "cmd.h"
#include <errno.h>
#if HAVE_SETJMP_H
# include <setjmp.h>
static jmp_buf			Env;
#endif

extern int			errno;

/*
 * Find the first command in Path that exists 
 */
extern char *CmdFind(Path)
     char		      **Path;
{
    register int		i;

    for (i = 0; Path[i]; ++i)
	if (access(Path[i], X_OK) == 0)
	    return Path[i];

    return (char *) NULL;
}

/*
 * Called when the alarm goes off
 */
static void TimeOutHandler(SigNum)
     int			SigNum;
{
    if (SigNum == SIGALRM)
	SImsg(SIM_DBG,
	      "TimeOutHandler: A time-out has occured executing a comamnd.");
    else
	SImsg(SIM_DBG,
	      "TimeOutHandler: Cmd Received signal %s",
	      SignalName(SigNum));

#if	defined(HAVE_LONGJMP)
    longjmp(Env, SigNum);
#endif	/* HAVE_LONGJMP */
}

/*
 * Open (create) a command.
 * Cmd->FD is the filedescriptor to read/write from/to.
 */
extern int CmdOpen(Cmd)
    Cmd_t		       *Cmd;
{
    pid_t			ProcID = 0;
    char		      **PtrPtr;
    int				IOpipe[2];
    int				ChildEnd;
    int				ChildStdEnd;
    int				ParentEnd;
    int				NullFD;
    int				WithPrivs = FALSE;
    char		       *cp;
    int 			LongJumpVal;

    if (!Cmd) {
	errno = EINVAL;
	return -1;
    }

    if (FLAGS_ON(Cmd->Flags, CMF_WITHPRIVS))
	WithPrivs = TRUE;

    Cmd->Program = CmdFind(Cmd->CmdPath);
    if (!Cmd->Program) {
	cp = strrchr(Cmd->Argv[0], '/');
	if (cp)
	    ++cp;
	else
	    cp = Cmd->Argv[0];
	SImsg(SIM_INFO, "%s: Cannot locate executable command.", cp);
	return -1;
    }
    Cmd->Argv[0] = Cmd->Program;

    if (Debug) {
	SImsg(SIM_INFO, "CmdOpen  <%s>", Cmd->Program);
	for (PtrPtr = Cmd->Argv+1; PtrPtr && *PtrPtr; ++PtrPtr)
	    SImsg(SIM_INFO, " <%s>", *PtrPtr);
	SImsg(SIM_INFO, "\t%s Privs\n", (WithPrivs) ? "With" : "Without");
    }

    /*
     * Create the I/O pipe
     */
    if (pipe(IOpipe) < 0) {
	SImsg(SIM_GERR, "Create pipe failed: %s", SYSERR);
	return -1;
    }

    /*
     * Setup descriptors
     */
    if (FLAGS_ON(Cmd->Flags, CMF_READ)) {
	ParentEnd = IOpipe[0];
	ChildEnd = IOpipe[1];
    } else if (FLAGS_ON(Cmd->Flags, CMF_WRITE)) {
	ParentEnd = IOpipe[1];
	ChildEnd = IOpipe[0];
    } else {
	SImsg(SIM_GERR, "CmdOpen() - Must specified CMF_READ or CMF_WRITE.");
	return -1;
    }

    ProcID = fork();

    if (ProcID < 0) {
	SImsg(SIM_GERR, "Fork failed, cannot run %s: %s", 
	      Cmd->Program, SYSERR);
	return -1;
    } else if (ProcID == 0) {
	/*
	 * Child
	 */

	(void) close(ParentEnd);
	ChildStdEnd = (FLAGS_ON(Cmd->Flags, CMF_READ)) ? 1 : 0;
	if (ChildEnd != ChildStdEnd)
	    if (dup2(ChildEnd, ChildStdEnd) < 0)
		SImsg(SIM_GERR, "dup2(%d, %d) failed: %s.", 
		      ChildEnd, ChildStdEnd, SYSERR);
	if (!Debug && FLAGS_OFF(Cmd->Flags, CMF_STDERR)) {
	    /*
	     * Redirect stderr to /dev/null
	     */
	    NullFD = open(_PATH_NULL, O_WRONLY);
	    (void) dup2(NullFD, fileno(stderr));
	}

#if	defined(HAVE_SETJMP)
	if (Cmd->TimeOut) {
	    /*
	     * Set a timeout since programs may hang
	     */
	    if ((LongJumpVal = setjmp(Env)) != 0) {
		/*
		 * If we reach this point then the time-out has occured
		 */
		Cmd->Flags |= CMF_TIMEDOUT;
		cp = SignalName(LongJumpVal);
		SImsg(SIM_DBG, "CmdOpen: Program terminated due to %s", 
		      cp);
		return -1;
	    }
	    TimeOutStart(Cmd->TimeOut, TimeOutHandler);
	}
#endif	/* HAVE_SETJMP */

	ExecInit(WithPrivs);
	execve(Cmd->Program, Cmd->Argv, Cmd->Env);

	if (Cmd->TimeOut)
	    TimeOutEnd();

	SImsg(SIM_GERR, "Execve \"%s\" failed: %s", Cmd->Program, SYSERR);
	exit(127);
    } else {
	/*
	 * Parent
	 */
	if (Cmd->TimeOut)
	    TimeOutEnd();
	(void) close(ChildEnd);

	Cmd->ProcID = ProcID;
	Cmd->FD = ParentEnd;

	return 0;
    }

    return -1;
}

/*
 * End a Cmd
 */
extern int CmdClose(Cmd)
     Cmd_t		       *Cmd;
{
    int				Status;

    if (!Cmd)
	return -1;

    (void) close(Cmd->FD);
    /*
     * Don't WaitForProc if there was a time out since that will likely
     * just hang us up again.
     */
    if (FLAGS_OFF(Cmd->Flags, CMF_TIMEDOUT)) {
	Status = WaitForProc(Cmd->ProcID);
	SImsg(SIM_DBG, "CmdClose <%s> exit status=%d", Cmd->Program, Status);
    } else {
	SImsg(SIM_DBG, "CmdClose <%s> not waiting due to command time out.",
	      Cmd->Program);
	Status = -1;
    }

    return(Status);
}
