/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.18 $";
#endif

/*
 * Device Find routines
 */

#include <stdio.h>
#include "defs.h"

#if 0
#define DEBUG_DEVFIND 	1	/* DDD/XXX Enable extra debug */
#endif

/*
 * Initialize a DevFind_t
 */
extern void DevFindInit(Find)
     DevFind_t		       *Find;
{
    if (!Find)
	return;

    memset(Find, CNULL, sizeof(DevFind_t));
    Find->DevType = Find->ClassType = Find->Unit = Find->Port = 
	Find->Bus = Find->DevNum = -1;
}

/*
 * Add a reason (What) to why Find matched
 */
static void DevFindAddMatch(Find, What, Item)
     DevFind_t		       *Find;
     char		       *What;
     char		       *Item;
{
    static char			Buff[512];

    if (!Find || !What)
	return;

    if ((strlen(Find->Reason) + strlen(What) + 2) >= sizeof(Find->Reason)) {
	SImsg(SIM_DBG, "DevFindAddMatch(..., %s): Find->Reason is full.", 
	      What);
	return;
    }

    if (Find->Reason && Find->Reason[0])
	StrAppend(Find->Reason, sizeof(Find->Reason), ",");

    StrAppend(Find->Reason, sizeof(Find->Reason), What);
    if (Item) {
	snprintf(Buff, sizeof(Buff), " [%s]", Item);
	StrAppend(Find->Reason, sizeof(Find->Reason), Buff);
    }	      
}

/*
 * Does NodeName match Tree?
 */
static int DevFindMatchName(Find)
     DevFind_t		       *Find;
{
    register char	      **cpp;
    char		       *NodeName;
    DevInfo_t		       *Tree;

    if (!Find || !Find->Tree)
	return(FALSE);

    Tree = Find->Tree;
    NodeName = Find->NodeName;

#if	DEBUG_DEVFIND
    SImsg(SIM_DBG, "DevFind: Look for node <%s> check <%s>",
	  PRTS(NodeName), PRTS(Tree->Name));
#endif

    if (Find->Expr == DFE_AND && !NodeName)
	return(TRUE);

    if (FLAGS_ON(Find->Flags, DFF_REGEX)) {
	if (NodeName && REmatch(strlower(Tree->Name), NodeName, NULL)) {
	    DevFindAddMatch(Find, "NodeName", NodeName);
	    return(TRUE);
	}
    } else {
	if (NodeName && EQ(NodeName, Tree->Name)) {
	    DevFindAddMatch(Find, "NodeName", NodeName);
	    return(TRUE);
	}
    }

    if (NodeName && Tree->Aliases)
	for (cpp = Tree->Aliases; cpp && *cpp; ++cpp) {
	    /* Skip alias if it's the same as the canonical name */
	    if (EQ(Tree->Name, *cpp))
		continue;
	    if (EQ(*cpp, NodeName)) {
		DevFindAddMatch(Find, "NodeNameAlias", NodeName);
		return(TRUE);
	    }
	}

    return(FALSE);
}

/*
 * Does Vendor match Tree?
 */
static int DevFindMatchVendor(Find)
     DevFind_t		       *Find;
{
    char		       *Name;
    DevInfo_t		       *Tree;

    if (!Find || !Find->Tree)
	return(FALSE);

    Tree = Find->Tree;
    Name = Find->Vendor;

    if (Find->Expr == DFE_AND && !Name)
	return(TRUE);

    if (FLAGS_ON(Find->Flags, DFF_REGEX)) {
	if (Name && REmatch(strlower(Tree->Vendor), Name, NULL)) {
	    DevFindAddMatch(Find, "Vendor", Name);
	    return(TRUE);
	}
    } else {
	if (Name && EQ(Name, Tree->Vendor)) {
	    DevFindAddMatch(Find, "Vendor", Name);
	    return(TRUE);
	}
    }

    return(FALSE);
}

/*
 * Does Serial match Tree?
 */
static int DevFindMatchSerial(Find)
     DevFind_t		       *Find;
{
    register Define_t	       *Def;
    static Define_t	       *IgnoreList;
    int				Hits = 0;

    if (!Find || !Find->Tree)
	return(FALSE);

    if (!Find->Serial) {
	if (Find->Expr == DFE_AND)
	    return TRUE;
	else
	    return FALSE;
    }

#define MIN_SERIAL_LEN 4
    if (strlen(Find->Serial) <= MIN_SERIAL_LEN) {
	SImsg(SIM_DBG,
	      "DevFindMatchSerial: %s: Serial <%s> is too short - IGNORING",
	      PRTS(Find->NodeName), PRTS(Find->Serial));
	return FALSE;
    }
	
    /*
     * If the Serial looks like a date (i.e. "12/01/99") then ignore it.
     */
    if (REmatch(Find->Serial, "[0-9][0-9]/[0-9][0-9]/[0-9][0-9]", NULL)) {
	SImsg(SIM_DBG, "DevFindMatchSerial: %s: Serial <%s> looks like a date instead of true serial# - IGNORING",
	      PRTS(Find->NodeName), PRTS(Find->Serial));
	return FALSE;
    }

    if (!IgnoreList)
	IgnoreList = DefGetList("IgnoreSerial");

    /*
     * Is Driver in the list of driver or serial names which we should ignore
     * Serial?
     */
    for (Def = IgnoreList; Def; Def = Def->Next) {
	Hits = 0;
	if (Def->ValStr1 && !EQ(Def->ValStr1, "-")) {
	    if (Find->Tree->Driver && 
		(EQ(Def->ValStr1, Find->Tree->Driver) || 
		 REmatch(Find->Tree->Driver, Def->ValStr1, NULL))) {
		Hits += 1;
	    } else {
		continue;
	    }
	}
	if (Def->ValStr2 && !EQ(Def->ValStr2, "-")) {
	    if (Find->Tree->Vendor && 
		(EQ(Def->ValStr2, Find->Tree->Vendor) ||
		 REmatch(Find->Tree->Vendor, Def->ValStr2, NULL))) {
		Hits += 1;
	    } else {
		continue;
	    }
	}
	if (Def->ValStr3 && !EQ(Def->ValStr3, "-")) {
	    if (Find->Tree->Model && 
		(EQ(Def->ValStr3, Find->Tree->Model) ||
		 REmatch(Find->Tree->Model, Def->ValStr3, NULL))) {
		Hits += 1;
	    } else {
		continue;
	    }
	}
	if (Def->ValStr4 && !EQ(Def->ValStr4, "-")) {
	    if (Find->Tree->Serial && 
		(EQ(Def->ValStr4, Find->Tree->Serial) ||
		 REmatch(Find->Tree->Serial, Def->ValStr4, NULL))) {
		Hits += 1;
	    } else {
		continue;
	    }
	}
	if (Hits) {
	    SImsg(SIM_DBG, 
		  "DevFindMatchSerial: ignoring serial# for %s; Matched IgnoreSerial %s (%d), Driver=%s",
		  PRTS (Find->Tree->Name), PRTS (Def->KeyStr), Def->KeyNum,
		  PRTS (Find->Tree->Driver));
	    return (FALSE);
	}
    }

    if (Find->Serial && *Find->Serial && 
	(EQ(Find->Serial, Find->Tree->Serial) || 
	 REmatch(Find->Tree->Serial, Find->Serial, NULL))) {
	DevFindAddMatch(Find, "Serial", Find->Serial);
	return(TRUE);
    } else
	return(FALSE);
}

/*
 * Does Ident match Tree?
 */
static int DevFindMatchIdent(Find)
     DevFind_t		       *Find;
{
    if (!Find || !Find->Tree)
	return(FALSE);

    if (Find->Expr == DFE_AND && !Find->Ident)
	return(TRUE);

    if (Find->Ident && IdentMatch(Find->Ident, Find->Tree->Ident)) {
	DevFindAddMatch(Find, "Ident", IdentString(Find->Ident));
	return(TRUE);
    } else
	return(FALSE);
}

/*
 * Does NodeID match Tree?
 */
static int DevFindMatchNodeID(Find)
     DevFind_t		       *Find;
{
    if (!Find || !Find->Tree)
	return(FALSE);

    if (Find->Expr == DFE_AND && Find->NodeID <= 0)
	return(TRUE);

    if (Find->NodeID && Find->NodeID != -1 && 
	Find->NodeID == Find->Tree->NodeID) {
	DevFindAddMatch(Find, "NodeID", itoa(Find->NodeID));
	return(TRUE);
    } else
	return(FALSE);
}

/*
 * Does Port match Tree?
 */
static int DevFindMatchPort(Find)
     DevFind_t		       *Find;
{
    if (!Find || !Find->Tree)
	return(FALSE);

    if (Find->Expr == DFE_AND && Find->Port == -1)
	return(TRUE);

    if (Find->Port != -1 && 
	Find->Port == Find->Tree->Port) {
	DevFindAddMatch(Find, "Port", itoa(Find->Port));
	return(TRUE);
    } else
	return(FALSE);
}

/*
 * Does Bus match Tree?
 */
static int DevFindMatchBus(Find)
     DevFind_t		       *Find;
{
    if (!Find || !Find->Tree)
	return(FALSE);

    if (Find->Expr == DFE_AND && Find->Bus == -1)
	return(TRUE);

    if (Find->Bus != -1 && 
	Find->Bus == Find->Tree->Bus) {
	DevFindAddMatch(Find, "Bus", itoa(Find->Bus));
	return(TRUE);
    } else
	return(FALSE);
}

/*
 * Does DevNum match Tree?
 */
static int DevFindMatchDevNum(Find)
     DevFind_t		       *Find;
{
    if (!Find || !Find->Tree)
	return(FALSE);

    if (Find->Expr == DFE_AND && Find->DevNum == -1)
	return(TRUE);

    if (Find->DevNum != -1 && 
	Find->DevNum == Find->Tree->DevNum) {
	DevFindAddMatch(Find, "DevNum", itoa(Find->DevNum));
	return(TRUE);
    } else
	return(FALSE);
}

/*
 * Does ClassType, DevType, and Unit match Tree?
 */
static int DevFindMatchType(Find)
     DevFind_t		       *Find;
{
    if (!Find || !Find->Tree)
	return FALSE;

    if (FLAGS_ON(Find->Flags, DFF_NOTYPE))
	return FALSE;

    /* Something has to be set or else we bail */
    if (Find->Unit < 0 && Find->DevType < 0 && Find->ClassType < 0) {
	if (Find->Expr == DFE_AND)
	    return TRUE;
	else
	    return FALSE;
    }

    if ((Find->Unit < 0 || (Find->Unit >= 0 && Find->Unit == Find->Tree->Unit)) &&
	(Find->DevType < 0 || Find->DevType == Find->Tree->Type) &&
	(Find->ClassType < 0 || Find->ClassType == Find->Tree->ClassType)) {
	SImsg(SIM_DBG, 
	      "DevFindMatchType: Dev=<%s> MATCHED Unit %d=%d Type %d=%d Class %d=%d",
	      PRTS(Find->NodeName),
	      Find->Unit, Find->Tree->Unit,
	      Find->DevType, Find->Tree->Type,
	      Find->ClassType, Find->Tree->ClassType);
	DevFindAddMatch(Find, "Type", NULL);
	return(TRUE);
    } else
	return(FALSE);
}

/*
 * --RECURSE--
 * Find a device in the device tree.  Find->Tree is the tree entry to search.
 * Find->Flags determines whether all conditions are OR'ed or AND'ed together. 
 * This function recursively calls itself.
 */
extern DevInfo_t *DevFind(Find)
     DevFind_t		       *Find;
{
    /*
     * Make sure local vars are not static since we recurse
     */
    DevInfo_t		       *Ptr;
    DevInfo_t	 	       *Tree;

    if (!Find || !Find->Tree)
	return((DevInfo_t *) NULL);

    /* Report possible loop */
    if (Find->Depth >= 5000 && Find->Depth % 1000 == 0) {
	SImsg(SIM_DBG, "DevFind: Recursion depth now %d - possible loop?",
	      Find->Depth);
    }
    /* Have we reached max depth? */
    if (Find->Depth >= 10000) {
	SImsg(SIM_DBG, "DevFind: Recursion depth max exceeded.");
	return((DevInfo_t *) NULL);
    }

    Tree = Find->Tree;

    if (Find->Expr == DFE_OR) {
	if (DevFindMatchName(Find) ||
	    DevFindMatchNodeID(Find) ||
	    DevFindMatchPort(Find) ||
	    DevFindMatchBus(Find) ||
	    DevFindMatchDevNum(Find) ||
	    DevFindMatchIdent(Find) ||
	    DevFindMatchSerial(Find) ||
	    DevFindMatchVendor(Find) ||
	    DevFindMatchType(Find)) {
#if	DEBUG_DEVFIND
	    SImsg(SIM_DBG, "DevFind: DFE_OR found node <%s>",
		  PRTS(Find->NodeName));
#endif
	    return(Tree);
	}
    } else if (Find->Expr == DFE_AND) {
	if (DevFindMatchName(Find) &&
	    DevFindMatchNodeID(Find) &&
	    DevFindMatchPort(Find) &&
	    DevFindMatchBus(Find) &&
	    DevFindMatchDevNum(Find) &&
	    DevFindMatchIdent(Find) &&
	    DevFindMatchSerial(Find) &&
	    DevFindMatchVendor(Find) &&
	    DevFindMatchType(Find)) {
#if	DEBUG_DEVFIND
	    SImsg(SIM_DBG, "DevFind: DFE_AND found node <%s>",
		  PRTS(Find->NodeName));
#endif
	    return(Tree);
	}
    } else {
	SImsg(SIM_DBG, "DevFind(): Expr %d unknown.", Find->Expr);
	return((DevInfo_t *) NULL);
    }

    /* Increment our recursion depth */
    ++Find->Depth;

    if (Tree->Slaves) {
	Find->Tree = Tree->Slaves;
	if (Ptr = DevFind(Find))
	    return(Ptr);
    }

    if (Tree->Next) {
	Find->Tree = Tree->Next;
	if (Ptr = DevFind(Find))
	    return(Ptr);
    }

    return((DevInfo_t *) NULL);
}
