/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.6 $";
#endif

/*
 * Linux USB functions
 */

#include "defs.h"
#include "strutil.h"

/*
 * The format of PATH_USB_DEVICES is defined in 
 * 	/usr/src/linux/drivers/usb/devices.c
 */
#if defined(USBDEBUG)
# define PATH_USB_DEVICES	"/home/mcooper/tmp/usb.devices"
#else
# define PATH_USB_DEVICES	"/proc/bus/usb/devices"
#endif

#define PARENT_DEVNUM_STR	"Parent Device #"

/*
 * Take Key=Value and return Value.
 */
static char *Value(StrArgv, Index)
     char		      **StrArgv;
     int			Index;
{
    static char		      **Argv;
    int				Argc;
    char		       *Val = NULL;
    char		       *cp;

    /*
     * Some entries are "Key=Value" while others are "Key=  Value" so 
     * we need to handle both cases.
     */
    Argc = StrToArgv(StrArgv[Index], "=", &Argv, NULL, 0);
    if ((Argc != 2) && StrArgv[Index+1] && !strchr(StrArgv[Index+1], '=')) {
	Val = StrArgv[Index+1];
    } else if (Argc != 2) {
	SImsg(SIM_DBG, "USB: Value malformed string <%s>", StrArgv[Index]);
	return (char *) NULL;
    } else if (Argc == 2) {
	Val = Argv[1];
    }

    /* Skip leading whitespace */
    for (cp = Val; cp && *cp && isspace(*cp); ++cp);

    return cp;
}

/*
 * Parse a "T" line and add relevent data to DevInfo.  Line looks like:
 * 	Bus=02 Lev=00 Prnt=00 Port=00 Cnt=00 Dev#=  1 Spd=12  MxCh= 2
 */
static int ParseT(Line, DevInfo)
     char		       *Line;
     DevInfo_t		       *DevInfo;
{
    static char		      **Argv;
    int				Argc;
    register int		i;
    static char			Buff[512];
    DevDesc_t		      **lp;
    char		       *cp;

    SImsg(SIM_DBG, "USB: Parsing T=<%s>", Line);

    lp = &(DevInfo->DescList);

    Argc = StrToArgv(Line, " ", &Argv, NULL, STRTO_SKIPRP);
    for (i = 0; i < Argc; ++i) {
	if (EQN(Argv[i], "Bus=", 4)) {
	    if (cp = Value(Argv, i))
		DevInfo->Bus = atoi(cp);
	} else if (EQN(Argv[i], "Lev=", 4)) {
	    if (cp = Value(Argv, i))
		AddDesc(lp, "Topology-tier Level", strdup(itoa(atoi(cp))), 0);
	} else if (EQN(Argv[i], "Port=", 5)) {
	    if (cp = Value(Argv, i))
		DevInfo->Port = atoi(cp);
	} else if (EQN(Argv[i], "Prnt=", 5)) {
	    if (cp = Value(Argv, i))
		AddDesc(lp, PARENT_DEVNUM_STR, strdup(itoa(atoi(cp))), 0);
	} else if (EQN(Argv[i], "Cnt=", 4)) {
	    if (cp = Value(Argv, i))
		AddDesc(lp, "Count", strdup(itoa(atoi(cp))), 0);
	} else if (EQN(Argv[i], "Dev#=", 5)) {
	    if (cp = Value(Argv, i))
		DevInfo->DevNum = atoi(cp);
	} else if (EQN(Argv[i], "Spd=", 4)) {
	    if (cp = Value(Argv, i)) {
		snprintf(Buff, sizeof(Buff), "%s Mb/s", cp);
		DevInfo->Speed = strdup(Buff);
	    }
	} else if (EQN(Argv[i], "MxCh=", 5)) {
	    if (cp = Value(Argv, i))
		AddDesc(lp, "Max Children", strdup(itoa(atoi(cp))), 0);
	}
    }

    return 0;
}

/*
 * Parse an "I" line and add relevent data to DevInfo.  Line looks like:
 * 	If#= 0 Alt= 0 #EPs= 2 Cls=07(print) Sub=01 Prot=02 Driver=usblp
 */
static int ParseI(Line, DevInfo)
     char		       *Line;
     DevInfo_t		       *DevInfo;
{
    static char		      **Argv;
    int				Argc;
    register int		i;
    static char			Buff[512];
    DevDesc_t		      **lp;
    char		       *cp;
    char		       *Val;

    SImsg(SIM_DBG, "USB: Parsing I=<%s>", Line);

    lp = &(DevInfo->DescList);

    Argc = StrToArgv(Line, " ", &Argv, NULL, STRTO_SKIPRP);
    for (i = 0; i < Argc; ++i) {
	if (EQN(Argv[i], "If#=", 4)) {
	    if (cp = Value(Argv, i))
		AddDesc(lp, "Interface #", strdup(itoa(atoi(cp))), 0);
	} else if (EQN(Argv[i], "Alt=", 4)) {
	    if (cp = Value(Argv, i))
		AddDesc(lp, "Alternative Setting", strdup(itoa(atoi(cp))), 0);
	} else if (EQN(Argv[i], "#EPs=", 5)) {
	    if (cp = Value(Argv, i))
		AddDesc(lp, "# End Points", strdup(itoa(atoi(cp))), 0);
	} else if (EQN(Argv[i], "Cls=", 4)) {
	    /*
	     * Value looks like "07(print)" or equivilent.
	     */
	    if (Val = Value(Argv, i)) {
		if (cp = strchr(Val, '(')) {
		    Val = cp + 1;
		    /* Find the start of the whitespace or ending ')' */
		    for (cp = Val; cp && *cp && isalpha(*cp); ++cp);
		    snprintf(Buff, sizeof(Buff), "%.*s", cp - Val, Val);
		    SImsg(SIM_DBG, "USB: Class name is <%s>", Buff);
		    /*
		     * See class_info[] in linux/drivers/usb/devices.c 
		     */
		    if (EQ(Buff, "print"))
			DevInfo->Type = DT_PRINTER;
		    else if (EQ(Buff, "stor."))
			DevInfo->Type = DT_DISKDRIVE;
		    else if (EQ(Buff, "hub"))
			DevInfo->Type = DT_BUS;
		    else if (EQ(Buff, "audio"))
			DevInfo->Type = DT_AUDIO;
		    else
			SImsg(SIM_DBG,
			      "USB: Class name <%s> is unknown to SysInfo",
			      Buff);
		}
	    }
	} else if (EQN(Argv[i], "Sub=", 4)) {
	    if (cp = Value(Argv, i))
		AddDesc(lp, "Sub Class", strdup(itoa(atoi(cp))), 0);
	} else if (EQN(Argv[i], "Prot=", 5)) {
	    if (cp = Value(Argv, i))
		AddDesc(lp, "Protocol", strdup(itoa(atoi(cp))), 0);
	} else if (EQN(Argv[i], "Driver=", 7)) {
	    if (cp = Value(Argv, i))
		AddDesc(lp, "Driver", strdup(cp), 0);
	}
    }

    return 0;
}

/*
 * Parse an "S" line and add relevent data to DevInfo.  Line looks like:
 *	Manufacturer=Lexmark International, Inc.
 *	Product=Lexmark Optra E310
 *	SerialNumber=0068255
 * 	
 */
static int ParseS(Line, DevInfo)
     char		       *Line;
     DevInfo_t		       *DevInfo;
{
    char		       *cp;

    SImsg(SIM_DBG, "USB: Parsing S=<%s>", Line);

    if (EQN(Line, "Manufacturer=", 13)) {
	if (cp = strchr(Line, '='))
	    DevInfo->Vendor = strdup(cp+1);
    } else if (EQN(Line, "Product=", 8)) {
	if (cp = strchr(Line, '='))
	    DevInfo->Model = strdup(cp+1);
    } else if (EQN(Line, "SerialNumber=", 13)) {
	if (cp = strchr(Line, '='))
	    DevInfo->Serial = strdup(cp+1);
    } else {
	SImsg(SIM_DBG, "USB: Unknown S: line entry: %s", Line);
    }

    /*
     * If the Model and Vendor both start with the same word, then that
     * means the Model includes part of the Vendor's name so skip over that
     * part so that DevInfo->Model doesn't contain the Vendor name.
     */
    if (DevInfo->Model && DevInfo->Vendor) {
	if (cp = strchr(DevInfo->Model, ' ')) {
	    if (EQN(DevInfo->Model, DevInfo->Vendor, cp - DevInfo->Model))
		DevInfo->Model = cp + 1;
	}
    }

    return 0;
}

/*
 * Create and add the DevInfo
 */
static int UsbDevInfoCreate(DevInfo_t *DevInfo, DevInfo_t **TreePtr,
			    char **SearchNames)
{
    char		       *Name = NULL;
    static char			Buff[128];
    DevDesc_t		       *dp;
    DevInfo_t		       *Parent = NULL;
    DevInfo_t		       *Dev = NULL;
    static DevFind_t		Find;
    int				ParDevNum = 0;

    DevInfo->ClassType = CT_USB;

    Name = NULL;
    switch (DevInfo->Type) {
    case DT_PRINTER:	Name = "lp";		break;
    case DT_AUDIO:	Name = "audio";		break;
    default:
	Name = "usb";
	DevInfo->Type = DT_SERIAL;
	break;
    }
    if (DevInfo->Unit < 0)
	DevInfo->Unit = UnitAssign(DevInfo->Type, -1, NULL);
    snprintf(Buff, sizeof(Buff), "%s%d", Name, DevInfo->Unit);
    DevInfo->Name = strdup(Buff);

    /*
     * See if there's an existing USB device which matches us
     * This will only be an approximation since any existing USB
     * devices are likely from DMI which doesn't provide the
     * bus/devnum data we need to match identically.
     */
    if (DevInfo->Type == DT_SERIAL && DevInfo->ClassType == CT_USB) {
	    DevFindInit(&Find);
	    Find.Expr = DFE_AND;
	    Find.Tree = *TreePtr;
	    Find.DevType = DT_SERIAL;
	    Find.ClassType = CT_USB;
	    Find.Unit = DevInfo->Unit;
	    if (Dev = DevFind(&Find))
		/* Graft the two together */
		DevGraft(DevInfo, Dev);
    } else if (TreePtr) {
	/*
	 * Try to find the parent of this device and associate it
	 */

	/* Find the parent's DevNum */
	for (dp = DevInfo->DescList; dp; dp = dp->Next)
	    if (EQ(dp->Label, PARENT_DEVNUM_STR)) {
		if (dp->Desc)
		    ParDevNum = atoi(dp->Desc);
		break;
	    }
	if (ParDevNum) {
	    DevFindInit(&Find);
	    Find.Expr = DFE_AND;
	    Find.Tree = *TreePtr;
	    Find.DevType = DT_SERIAL;
	    Find.ClassType = CT_USB;
	    Find.Bus = DevInfo->Bus;
	    Find.DevNum = ParDevNum;
	    Parent = DevFind(&Find);
	}
    }
    if (Parent) {
	DevInfo->Master = Parent;
	DevInfo->MasterName = Parent->Name;
    }

    AddDevice(DevInfo, TreePtr, SearchNames);

    return 0;
}

/*
 * Detect devices attached to USB
 */
extern int DetectUsb(DevInfo_t **TreePtr, char **SearchNames)
{
    DevInfo_t		       *DevInfo = NULL;
    static char			Buff[1024];
    char		       *cp;
    char		       *Name = NULL;
    FILE		       *FH;

    SImsg(SIM_DBG, "USB: Detecting devices ...");

    FH = fopen(PATH_USB_DEVICES, "r");
    if (!FH) {
	SImsg(SIM_DBG, "USB: %s: Open for read failed: %s",
	      PATH_USB_DEVICES, SYSERR);
	return -1;
    }

    while (fgets(Buff, sizeof(Buff), FH)) {
	if (cp = strchr(Buff, '\n'))
	    *cp = CNULL;

	SImsg(SIM_DBG, "USB: LINE=%s", Buff);

	/* Skip cp over first argument and position at start of 2nd arg */
	cp = strchr(Buff, ' ');
	if (!cp)
	    cp = Buff;
	for ( ; cp && *cp && isspace(*cp); ++cp);

	if (EQN(Buff, "T:", 2)) {
	    /* Start of a new entry */
	    if (DevInfo)
		UsbDevInfoCreate(DevInfo, TreePtr, SearchNames);
	    DevInfo = DevInfoNew(NULL);
	    ParseT(cp, DevInfo);
	} else if (EQN(Buff, "I:", 2)) {
	    ParseI(cp, DevInfo);
	} else if (EQN(Buff, "S:", 2)) {
	    ParseS(cp, DevInfo);
	}
    }
    fclose(FH);

    /* Create the last entry, if any */
    if (DevInfo)
	UsbDevInfoCreate(DevInfo, TreePtr, SearchNames);

    return 0;
}
