/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.3 $";
#endif

#include "defs.h"
#include "dmi.h"

DevInfo_t	 	       *BiosDevInfo = NULL;
static char		       *s = "Supports";
static DevDesc_t	      **lp;

/*
 * Create a new DevInfo BIOS device
 */
extern DevInfo_t *NewBiosDevInfo()
{
    DevInfo_t		       *DevInfo;
    static char			Buff[128];
    static int			Unit;

    DevInfo = NewDevInfo(NULL);
    DevInfo->Unit = Unit++;
    snprintf(Buff, sizeof(Buff), "bios%d", DevInfo->Unit);
    DevInfo->Name = strdup(Buff);
    DevInfo->Type = DT_BIOS;

    BiosDevInfo = DevInfo;

    return DevInfo;
}

/*
 * Decode DMI BIOS (DMI_STRUCT_BIOS)
 */
extern DevInfo_t *DmiDecodeBios(Hdr, Data, Dmi, Query)
     DmiHeader_t	       *Hdr;
     void		       *Data;
     Dmi_t		       *Dmi;
     DmiQuery_t		       *Query;
{
    DevInfo_t		       *DevInfo;
    DmiStructBios_t	       *Bios;
    char		       *cp;
    static char			Buff[128];

    Bios = (DmiStructBios_t *) Data;

    /*
     * Use the existing BIOS DevInfo if found, else create it
     */
    if (BiosDevInfo)
	DevInfo = BiosDevInfo;
    else
	DevInfo = NewBiosDevInfo();

    lp = &(DevInfo->DescList);

    DevInfo->Vendor = DmiString(Hdr, Bios->Vendor);
    /* 
     * Looks like many BIOS treat Model==Version
     */
    DevInfo->Model = DmiString(Hdr, Bios->Version);
    if (cp = DmiString(Hdr, Bios->ReleaseDate))
	AddDesc(lp, "Release Date", cp, 0);	
    
    SImsg(SIM_DBG, "DMI BIOS: Vendor=<%s> Model=<%s>",
	  PRTS(DevInfo->Vendor), PRTS(DevInfo->Model));

    /*
     * If this is a specific query, then set the values now and return
     */
    if (Query && Query->Type == Hdr->Type) {
	Query->ValStr1 = DevInfo->Vendor;
	Query->ValStr2 = DevInfo->Model;
	return DevInfo;
    }

    /*
     * Set general DMI values
     */
    if (Dmi->SmbiosMajVer) {
	snprintf(Buff, sizeof(Buff), "%d.%d",
		 Dmi->SmbiosMajVer, Dmi->SmbiosMinVer);
	AddDesc(lp, "SMBIOS Version", strdup(Buff), 0);
    }
    if (Dmi->SmbiosEpsRev)
	AddDesc(lp, "SMBIOS EPS Revision", itoa(Dmi->SmbiosEpsRev), 0);
    if (Dmi->PnpBios || Bios->PnP)	
	AddDesc(lp, s, "Plug and Play", 0);
    if (Dmi->SysId)
	AddDesc(lp, s, "SYSID", 0);

    /*
     * Set BIOS specific values
     */
    if (Bios->StartAddr)
	/* Per SMBIOS Ref Spec 3.3.1 */
	if (cp = GetSizeStr((Large_t)((0x10000 - Bios->StartAddr) * 16),BYTES))
	    AddDesc(lp, "Runtime Size", cp, 0);
    if (Bios->RomSize)
	if (cp = GetSizeStr((Large_t)(Bios->RomSize * 64), KBYTES))
	    AddDesc(lp, "ROM Size", cp, 0);
    if (Bios->ISA)	AddDesc(lp, s, "ISA", 0);
    if (Bios->MCA)	AddDesc(lp, s, "MCA", 0);
    if (Bios->EISA)	AddDesc(lp, s, "EISA", 0);
    if (Bios->PCI)	AddDesc(lp, s, "PCI", 0);
    if (Bios->AGP)	AddDesc(lp, s, "AGP", 0);
    if (Bios->PCCard)	AddDesc(lp, s, "PC Card", 0);
    if (Bios->APM)	AddDesc(lp, s, "APM", 0);
    if (Bios->FlashUpgrade) AddDesc(lp, s, "Flash Upgradeable", 0);
    if (Bios->Shadow)	AddDesc(lp, s, "Shadowing allowed", 0);
    if (Bios->VLVESA)	AddDesc(lp, s, "VL-VESA", 0);
    if (Bios->ESCD)	AddDesc(lp, s, "ESCD", 0);
    if (Bios->ACPI)	AddDesc(lp, s, "ACPI", 0);
    if (Bios->SelectBoot) AddDesc(lp, s, "Selectable Boot", 0);
    if (Bios->CDboot)	AddDesc(lp, s, "Boot from CD", 0);
    if (Bios->PCCardBoot)AddDesc(lp, s, "Boot from PC Card", 0);
    if (Bios->I2Oboot)	AddDesc(lp, s, "Boot from I2O", 0);
    if (Bios->LS120boot)AddDesc(lp, s, "Boot from LS-120", 0);
    if (Bios->AtapiZipBoot)AddDesc(lp, s, "Boot from ATAPI ZIP", 0);
    if (Bios->ROMsocketed) AddDesc(lp, s, "ROM is socketed", 0);
    if (Bios->Flp5in360k) AddDesc(lp, s, "Floppy 5.25\" 360KB", 0);
    if (Bios->Flp5in12m) AddDesc(lp, s, "Floppy 5.25\" 1.2MB", 0);
    if (Bios->Flp3in720k) AddDesc(lp, s, "Floppy 3.5\" 720KB", 0);
    if (Bios->Flp3in288m) AddDesc(lp, s, "Floppy 3.5\" 2.88MB", 0);
    if (Bios->JpnNEC9800) 
	AddDesc(lp, s, "Int 13h - Japanense Floppy NEC 9800 1.2mb", 0);
    if (Bios->JpnToshiba) 
	AddDesc(lp, s, "Int 13h - Japanense Floppy Toshiba 1.2mb", 0);
    if (Bios->Serial)	AddDesc(lp, s, "Int 14h Serial Services", 0);
    if (Bios->Printer)	AddDesc(lp, s, "Int 17h Printer Services", 0);
    if (Bios->CGA)	AddDesc(lp, s, "Int 10h CGA/Mono Video Services", 0);
    if (Bios->NecPC98)	AddDesc(lp, s, "NEC PC-98", 0);
    if (Bios->USBlegacy) AddDesc(lp, s, "USB Legacy", 0);
    if (Bios->SmartBattery) AddDesc(lp, s, "Smart Battery", 0);
    if (Bios->FuncKeyNetBoot)
	AddDesc(lp, s, "Function Key Initiated Network Boot", 0);
    if (Bios->EDD)	AddDesc(lp, s, "Enhanced Disk Drive (EDD)", 0);
    if (Bios->PrintScreen) AddDesc(lp, s, "Print Screen Service", 0);
    if (Bios->Kbd8042)	AddDesc(lp, s, "8042 keyboard", 0);
    
    return DevInfo;
}
