/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.3 $";
#endif

#include "defs.h"
#include "dmi.h"

static char		       *s = "Supports";
static DevDesc_t	      **lp;

/*
 * Values for System Enclosure's (DmiStructEnclosure_t.Type)
 * See section 3.3.4.1 of SMBIOS Ref Spec
 */
static char *EnclosureTypes[] = {
    "",
    "",
    "",
    "Desktop",
    "Low Profile Desktop",
    "Pizza Box",
    "Mini Tower",
    "Tower",
    "Portable",
    "Laptop",
    "Notebook",
    "Handheld",
    "Docking Station",
    "All in One",
    "Sub Notebook",
    "Space-saving",
    "Lunch Box",
    "Main Server Chassis",
    "Expansion Chassis",
    "SubChassis",
    "Bus Expansion Chassis",
    "Peripheral Chassis",
    "RAID Chassis",
    "Rack Mount Chassis",
    "Sealed-case PC",
    NULL
};

/*
 * Values for System Enclosure State's (DmiStructEnclosure_t)
 * See section 3.3.4.2 of SMBIOS Ref Spec.
 */
static char *EnclosureStates[] = {
    "",
    "",
    "Safe",
    "Warning",
    "Critical",
    "Non-recoverable",
    NULL
};

/*
 * Values for System Enclosure Security State's (DmiStructEnclosure_t)
 * See section 3.3.4.2 of SMBIOS Ref Spec.
 */
static char *EnclosureSecurityStates[] = {
    "",
    "",
    "",
    "External interface locked out",
    "External interface enabled",
    NULL
};

/*
 * Decode DMI System Information (DMI_STRUCT_ENCLOSURE)
 */
extern DevInfo_t *DmiDecodeEnclosure(Hdr, Data, Dmi, Query)
     DmiHeader_t	       *Hdr;
     void		       *Data;
     Dmi_t		       *Dmi;
     DmiQuery_t		       *Query;
{
    DevInfo_t		       *DevInfo;
    DmiStructEnclosure_t       *Enc;
    char		       *Mfg = NULL;
    char		       *Model = NULL;
    char		       *Version = NULL;
    char		       *Serial = NULL;
    char		       *Asset = NULL;
    char		       *cp;
    static char			Buff[128];
    static int			Unit;

    Enc = (DmiStructEnclosure_t *) Data;

    if ((cp = DmiString(Hdr, Enc->Mfg)) && !DmiIsGeneric(cp))
	Mfg = cp;
    if (cp = GetArrayEntry(EnclosureTypes, Enc->Type))
	Model = cp;
    if ((cp = DmiString(Hdr, Enc->Version)) && !DmiIsGeneric(cp))
	Version = cp;
    if ((cp = DmiString(Hdr, Enc->Serial)) && !DmiIsGeneric(cp))
	Serial = cp;
    if ((cp = DmiString(Hdr, Enc->AssetTag)) && !DmiIsGeneric(cp))
	Asset = cp;

    SImsg(SIM_DBG, 
	  "DMI ENCLOSURE: Mfg=<%s> Type=<%s> Version=<%s> Serial=<%s>",
	  PRTS(Mfg), PRTS(Model), PRTS(Version), PRTS(Serial));

    /*
     * If this is a specific query, then set the values now and return
     */
    if (Query && Query->Type == Hdr->Type) {
	Query->ValStr1 = Mfg;
	Query->ValStr2 = Model;
	Query->ValStr3 = Version;
	Query->ValStr4 = Serial;
	Query->ValStr5 = Asset;
	return DevInfo;
    }

    DevInfo = NewDevInfo(NULL);
    DevInfo->Unit = Unit++;
    snprintf(Buff, sizeof(Buff), "enclosure%d", DevInfo->Unit);
    DevInfo->Name = strdup(Buff);
    DevInfo->Type = DT_ENCLOSURE;

    lp = &(DevInfo->DescList);

    if (Mfg)		DevInfo->Vendor = Mfg;
    if (Model)		DevInfo->Model = Model;
    if (Version)	DevInfo->Revision = Version;
    if (Serial)		DevInfo->Serial = Serial;
    if (Asset)
	AddDesc(lp, "Asset Tag", Asset, 0);
    if (Enc->Lockable)	AddDesc(lp, s, "Is lockable", 0);

    if (Dmi->SmbiosVersion >= 2010) {
	if (cp = GetArrayEntry(EnclosureStates, Enc->BootUpState))
	    AddDesc(lp, "Boot Up State", cp, 0);
	if (cp = GetArrayEntry(EnclosureStates, Enc->PwrSupplyState))
	    AddDesc(lp, "Power Supply State", cp, 0);
	if (cp = GetArrayEntry(EnclosureStates, Enc->ThermalState))
	    AddDesc(lp, "Thermal State", cp, 0);
	if (cp = GetArrayEntry(EnclosureSecurityStates, Enc->SecurityStatus))
	    AddDesc(lp, "Security State", cp, 0);
    }

    return DevInfo;
}
