/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.32 $";
#endif

/*
 * DOS Partition functions
 */

#include "defs.h"
#include "dospart.h"

/*
 * Private declares
 */
static int		GetExtended();
static int		GetParts();

/*
 * Read from fd into Buffer.  Same as read(), but reads the full
 * Amount or until EOF.
 */
extern size_t Read(fd, Buffer, Amount)
     int			fd;
     void		       *Buffer;
     size_t			Amount;
{
    void		       *BufPtr;
    size_t			ReadAmt;
    size_t			Remainer;

    BufPtr = Buffer;
    Remainer = Amount;
    while (ReadAmt = read(fd, BufPtr, Remainer)) {
	if (ReadAmt < 0)
	    return ReadAmt;

	Remainer -= ReadAmt;
	BufPtr = (void *) ((char *)BufPtr + ReadAmt);

	if (Remainer <= 0)
	    return Amount;
    }

    return -1;
}

/*
 * Read the DOS Partition Table from File at Location
 * Return < 0 on error.
 * Return 0 on success.
 * Return 1 if no partition.
 */
static int ReadTable(Table, Sector, File, fd)
     DosTable_t		       *Table;
     Large_t			Sector;
     char		       *File;
     int			fd;
{
    int				Amt;
    int				SecSize = DOS_SECT_SIZE;
    Offset_t			Offset;
    Offset_t			RetVal;

    Offset = (Offset_t)((Offset_t)Sector * (Offset_t)SecSize);

#if	HAVE_PRINTF_LL
    SImsg(SIM_DBG, "%s: Looking for DOS Table at sector %lld Offset=%lld",
	  File, Sector, (Large_t)Offset);
#else
    SImsg(SIM_DBG, "%s: Looking for DOS Table at sector %ld Offset=%ld",
	  File, (long)Sector, (long)Offset);
#endif

    /*
     * This silly call is required on at least Red Hat 7.1 (Linux 2.4.2-2)
     * in order to get consistant success out of the following, real seek.
     */
    RetVal = myllseek(fd, (Offset_t)0, SEEK_SET);
    if (RetVal == (Offset_t)-1)
	   SImsg(SIM_DBG, "%s: seek reset to 0 failed: %s", File, SYSERR);

    /*
     * Do the real seek to Offset
     */
    RetVal = myllseek(fd, Offset, SEEK_SET);
    if (RetVal == (Offset_t)-1) {
#if	HAVE_PRINTF_LL
	SImsg(SIM_GERR, "%s: seek to %lld failed: %s", File, Offset, SYSERR);
#else
	SImsg(SIM_GERR, "%s: seek to %ld failed: %s", File, (long)Offset,
	      SYSERR);
#endif
	return -1;
    }

    for (SecSize = MIN_SEC_SIZE; SecSize <= MAX_SEC_SIZE; SecSize *= 2) {
	Amt = Read(fd, (void *) Table->Buffer.Buffer, (size_t)SecSize);
	if (Amt == SecSize) {
	    /* Success */
	    if (Table->Data.Signature != DOS_BOOT_MAGIC) {
		SImsg(SIM_DBG, "%s: No DOS Partition Table found.", File);
		return(1);
	    }
	    SImsg(SIM_DBG, "%s: DOS Table found with SecSize=%d",
		  File, SecSize);
	    return(0);
	}
    }

    SImsg(SIM_GERR, "%s: read DOS Table failed.", File);

    return(-1);
}

/*
 * Return description of what Dos Partition Type is
 */
extern char *DosPartType(Type)
     u_char			Type;
{
    Define_t		       *Def;

    if (Def = DefGet(DL_DOSPARTTYPES, NULL, (long)Type, 0))
	/* Return the short name else the long name */
	return (Def->ValStr2) ? Def->ValStr2 : Def->ValStr1 ;

    SImsg(SIM_DBG, "DOS Partition Type 0x%2X (%d) is unknown.", Type, Type);

    return((char *) NULL);
}

/*
 * Set type information in PartInfo using Type
 */
static int DosPartSetType(PartInfo, Type)
     PartInfo_t		       *PartInfo;
     u_char			Type;
{
    Define_t		       *Def;
    static char			Buff[64];

    PartInfo->TypeNum = Type;

    if (Def = DefGet(DL_DOSPARTTYPES, NULL, (long)Type, 0)) {
	if (Def->ValStr2 && !EQ(Def->ValStr2, "unused"))
	    PartInfo->Type = Def->ValStr2;
	if (Def->ValStr1 && !EQ(Def->ValStr1, "unused"))
	    PartInfo->TypeDesc = Def->ValStr1;
	if (!PartInfo->Type && !PartInfo->TypeDesc && Type) {
	    (void) snprintf(Buff, sizeof(Buff), "Unknown Type 0x%02X (%d)",
			    Type, Type);
	    PartInfo->TypeDesc = strdup(Buff);
	}
	return 0;
    } else {
	SImsg(SIM_DBG, "DOS Partition Type 0x%2X (%d) is unknown.", 
	      Type, Type);
	return -1;
    }
}

/*
 * Get standard DOS partition information from DosParts (a list) and add
 * the results to DiskDrive.
 *
 * Extended partitions are "linked" together.  When you are reading
 * an extended partition table and encounter a partition (usually the
 * 2nd partition) of type EXTENDED, then that entry indicates where the
 * next (logical) partition exists.  In such a case, the linked partition
 * is read from disk.  In effect the start of each extended logical partition
 * contains a DosTable_t table which has 2 entries.  The first contains
 * the data for the current logical partition.  The second contains "pointer"
 * data to where the next DosTable_t is on disk (if any).
 *
 * This function is called recursively whenever an EXTENDED partition is
 * found.
 *
 * --RECURSE--
 */
static int GetParts(DosParts, AbsSector, DiskDrive, DevFile, fd,
		    PartNum, Level)
     DosPart_t		       *DosParts;
     Large_t			AbsSector;	/* Absolute sector sector */
     DiskDrive_t	       *DiskDrive;
     char		       *DevFile;
     int			fd;
     int			PartNum;	/* Start Part #'s with this */
     int			Level;
{
    DiskPart_t		       *NewPart;
    static DosTable_t		Table;
    static DiskPart_t	       *Last;
    static PartInfo_t	       *PartInfo;
    static Large_t		ExtStartSect;
    static Large_t		StartSect;
    static char			Path[128];
    static char			Slice[32];
    static int			i;
    int				pn;
    int				Num;
    char		       *cp;
    char		      **cpp;

    SImsg(SIM_DBG,
#if	HAVE_PRINTF_LL
	  "%s: DOSGetParts AbsSector=%lld, PartNum=%d Level=%d",
	  DevFile, AbsSector,
#else
    	  "%s: DOSGetParts AbsSector=%ld, PartNum=%d Level=%d",
	  DevFile, (long)AbsSector,
#endif
	  PartNum, Level);

    Last = NULL;
    ExtStartSect = 0;
    StartSect = AbsSector;

    for (i = 0, pn = 0; i < DOS_NPARTS; ++i) {
	StartSect += (Large_t) DosParts[i].dp_start;
	if (IS_EXTENDED(DosParts[i].dp_type)) {
	    /*
	     * Remember this for when we're ready 
	     * The DOS partition scheme is to list all primary
	     * partitions first, then list the extended partitions.
	     */
	    if (IS_PRIMARY(PartNum))
		ExtStartSect = (Large_t) DosParts[i].dp_start + AbsSector;
	    else
		ExtStartSect = (Large_t) DosParts[i-1].dp_start + AbsSector +
		    (Large_t) DosParts[i-1].dp_size;
	}

	SImsg(SIM_DBG,
	      "%s:     PartNum=%d i=%d  Extended=%d",
	      DevFile, PartNum, i, IS_EXTENDED(DosParts[i].dp_type));
	SImsg(SIM_DBG,
#if	HAVE_PRINTF_LL
	      "%s:             StartSect=%lld ExtStartSect=%lld Type=0x%02x",
	      DevFile, StartSect, ExtStartSect,
#else
	      "%s:             StartSect=%ld ExtStartSect=%ld Type=0x%02x",
	      DevFile, (long)StartSect, (long)ExtStartSect,
#endif	/* HAVE_PRINTF_LL */
	      DosParts[i].dp_type);
	SImsg(SIM_DBG, 
	      "%s:             DP: Start=%ld Size=%ld StartCyl=%d EndCyl=%d",
	      DevFile,
	      DosParts[i].dp_start, DosParts[i].dp_size,
	      DosCylinder(DosParts[i].dp_ssect, DosParts[i].dp_scyl),
	      DosCylinder(DosParts[i].dp_esect, DosParts[i].dp_ecyl)
	      );

	/*
	 * Does this Primary partition contain an Extended partition?
	 * Nested Extended Partitions are not allowed.
	 */
	if (IS_EXTENDED(DosParts[i].dp_type)) {
	    SImsg(SIM_DBG, 
		  "%s:             Found Extended Part",
		  DevFile);
	    if (Level > 0)
		/*
		 * We've already found the top level Extended part so
		 * our job now is just to follow the linked partition
		 * trail.
		 */
		break;
	}

	if (DosParts[i].dp_size == 0 && DosParts[i].dp_type == 0) {
	    if (Level == 0)
		PartNum++;
	    SImsg(SIM_DBG, 
		  "%s:             Skipping since it's empty/unused.",
		  DevFile);
	    continue;
	}

	/*
	 * Time to put it all together and add the partition
	 */
	(void) snprintf(Slice, sizeof(Slice), "s%d", PartNum);
	(void) snprintf(Path, sizeof(Path), "%s%s", DevFile, Slice);

	PartInfo = PartInfoCreate(NULL);
	PartInfo->Num = PartNum;
	PartInfo->DevPath = strdup(Path);
	PartInfo->Name = strdup(Slice);
	PartInfo->SecSize = DiskDrive->SecSize;

	if (i == 0)
	    PartInfo->Title = "DOS";

	/*
	 * We use Type and Usage backwards here because Usage gets fully
	 * displayed on output and Type gets truncated.
	 */
	if (DosParts[i].dp_flags == DOS_ACTIVE) {
	    cpp = (char **) xcalloc(2, sizeof(char *));
	    cpp[0] = strdup("ACTIVE");
	    PartInfo->MntOpts = cpp;
	}

	(void) DosPartSetType(PartInfo, DosParts[i].dp_type);
	PartInfo->TypeNum = DosParts[i].dp_type;

	PartInfo->StartSect = StartSect;
	PartInfo->NumSect = (Large_t) DosParts[i].dp_size;

	if (IS_EXTENDED(DosParts[i].dp_type))
	    PartInfo->Usage = PIU_UNKNOWN;

	/*
	 * Add it to the list
	 */
	NewPart = NewDiskPart(NULL);
	NewPart->PartInfo = PartInfo;
	/*
	 * If DiskDrive exists, find the end of the list and set Last
	 */
	if (!Last && DiskDrive->DiskPart)
	    for (Last = DiskDrive->DiskPart; Last && Last->Next; 
		 Last = Last->Next);
	if (Last) {
	    Last->Next = NewPart;
	    Last = NewPart;
	} else {
	    DiskDrive->DiskPart = Last = NewPart;
	}
	SImsg(SIM_DBG, 
	      "%s:             Added Partition", DevFile);

	PartNum++;
    }

    if (ExtStartSect) {  
	if (ReadTable(&Table, ExtStartSect, DevFile, fd) != 0)
	    return -1;

	/*
	 * Now do standard Partition extraction
	 */
	(void) GetParts(&Table.Data.Parts[0], ExtStartSect, DiskDrive,
			DevFile, fd, 
			PartNum,
			Level+1);
    }

    return 0;
}

/*
 * Read all the DOS Parts for DevInfo
 * DevFile should be the name of the control device to read.
 */
extern int DosPartGet(Query)
     DosPartQuery_t	       *Query;
{
    DevInfo_t		       *DevInfo;
    char		       *CtlDev;
    char		       *BaseFile;
    char		       *DevFile;
    int				Status;
    static DosTable_t		Table;
    DiskDriveData_t	       *DiskDriveData;
    DiskDrive_t		       *Disk;
    PartInfo_t		       *PartInfo;
    int				fd;

    if (!Query)
	return -1;

    DevInfo = Query->DevInfo;
    CtlDev = Query->CtlDev;
    BaseFile = Query->BasePath;

    SImsg(SIM_DBG, "DosPartGet: CtlDev=<%s> BasePath=<%s> Flags=0x%x",
	  PRTS(Query->CtlDev), PRTS(Query->BasePath), Query->Flags);

    /*
     * Insure we only try to read DOS tables from disks.
     */
    if (DevInfo->Type != DT_DISKDRIVE) {
    	SImsg(SIM_DBG, "DosPartGet: Device Type (%s) is not DISKDRIVE.",
	      PRTS(TypeGetType(DevInfo->Type)));
	return -1;
    }
	
    if (!DevInfo || !DevInfo->Name || !CtlDev)
	return(-1);

    if (!BaseFile)
	BaseFile = CtlDev;

    /*
     * Open it ourself to get the right open flags.
     */
    DevFile = BaseFile;
    fd = open(BaseFile, O_RDONLY|O_NDELAY|O_NONBLOCK);
    if (fd < 0) {
        SImsg(SIM_GERR, "%s: DosPartGet: Open for read failed: %s", BaseFile, SYSERR);
	/*
	 * Try the CtlDev as backup
	 */
        SImsg(SIM_DBG, "%s: DosPartGet: Trying to open CtlDev=<%s>", BaseFile, CtlDev);
	fd = open(CtlDev, O_RDONLY|O_NDELAY|O_NONBLOCK);
	if (fd < 0) {
	    SImsg(SIM_GERR, "%s: DosPartGet: Open for read failed (2nd try): %s", CtlDev, SYSERR);
	    return(-1);
	}
	DevFile = CtlDev;
    }

    Status = ReadTable(&Table, (Large_t)DOS_BB_SECTOR, DevFile, fd);
    if (Status < 0) {
	close(fd);
	return(-1);
    } else if (Status > 0) {
	close(fd);
	return(0);
    }

    if (BaseFile != CtlDev)
	DevAddFile(DevInfo, strdup(BaseFile));
    DevAddFile(DevInfo, strdup(CtlDev));

    if (DevInfo->DevSpec)
	DiskDriveData = (DiskDriveData_t *) DevInfo->DevSpec;
    else {
	DiskDriveData = NewDiskDriveData(NULL);
	DevInfo->DevSpec = (void *) DiskDriveData;
    }

    if (DiskDriveData->HWdata)
	Disk = DiskDriveData->HWdata;
    else {
	Disk = NewDiskDrive(NULL);
	DiskDriveData->HWdata = Disk;
    }

    /* DOS Partition numbers start with 1 */
    GetParts(&Table.Data.Parts[0], (Large_t)0, Disk, DevFile, fd, 1, 0);

    (void) close(fd);

    return(0);
}
