/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.9 $";
#endif

/*
 * System Requirements
 */
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "autoconfig.h"
#include "defs.h"

#if 	defined(NEED_ISDIR)
/*
 * Is Path a directory?
 */
extern int IsDir(Path)
     char		       *Path;
{
    struct stat			Stat;

    if (stat(Path, &Stat) == 0 && S_ISDIR(Stat.st_mode))
	return 1;

    return 0;
}
#endif	/* NEED_ISDIR */

/*
 * Make directories required for File.
 */
extern int MakeFileDir(File)
     char		       *File;
{
    char		       *Dir;
    char		       *cp;

    /*
     * Make sure we have all directories leading up to Dst
     */
    Dir = File;
    if (cp = strrchr(Dir, '/')) {
	*cp = CNULL;
	if (MakeDir(Dir, 0755) < 0) {
	    *cp = '/';
	    Error("%s: Cannot create directories for installing %s: %s",
		  Dir, File, SYSERR);
	    return -1;
	}
	*cp = '/';
    }

    return 0;
}

/*
 * Copy a file
 */
extern int Copy(Src, Dst)
     char		       *Src;
     char		       *Dst;
{
    static char		        Buff[BUFSIZ];
    static char		        Path[MAXPATHLEN];
    char		       *bp;
    size_t			ReadAmt;
    size_t			WriteAmt;
    size_t			TotalAmt;
    int				SFD;
    int				DFD;
    int				Status = 0;
    register char	       *cp;

    if (!Src || !Dst)
	return -1;

    Debug("Copy <%s> -> <%s>", Src, Dst);

    SFD = open(Src, O_RDONLY);
    if (SFD < 0) {
	Error("%s: Open for read failed: %s", Src, SYSERR);
	return -1;
    }

    /*
     * If Dst is a directory then add the base of Src to Dst
     */
    if (IsDir(Dst)) {
	if (cp = strrchr(Src, '/')) {
	    (void) snprintf(Path, sizeof(Path), "%s/%s", Dst, cp+1);;
	    Dst = Path;
	}
    }

    if (MakeFileDir(Dst) < 0) {
	close(SFD);
	return -1;
    }

    DFD = open(Dst, O_WRONLY|O_CREAT, 0);
    if (DFD < 0) {
	Error("%s: Open for write/create failed: %s", Dst, SYSERR);
	close(SFD);
	return -1;
    }

    while ((ReadAmt = read(SFD, Buff, sizeof(Buff))) > 0) {
	for (TotalAmt = 0, bp = Buff; TotalAmt < ReadAmt; ) {
	    WriteAmt = write(DFD, bp, ReadAmt - (bp - Buff)); 
	    TotalAmt += WriteAmt;
	    bp += WriteAmt;
	}
    }

#ifdef notdef
    /* Should we somehow check for error on last read?  ReadAmt is unsigned
     * so we can't check < 0 */
    if (ReadAmt < 0) {
	Error("%s: Read failed: %", Src, SYSERR);
	Status = -1;
    }
#endif

    close(SFD);
    close(DFD);

    return Status;
}

/*
 * Return the Value of an edit Key
 */
extern char *EditValue(Key, Info, File)
     char		       *Key;
     McdInfo_t		       *Info;
     char		       *File;
{
    static char			Path[MAXPATHLEN];

    Path[0] = CNULL;
    if (EQ(Key, "prefix"))
	return Info->InstallDir;
    else if (EQ(Key, "cfdir"))
	(void) snprintf(Path, sizeof(Path), "%s/config", Info->InstallDir);
    else if (EQ(Key, "libexecdir"))
	(void) snprintf(Path, sizeof(Path), "%s/libexec", Info->InstallDir);
    else {
	Debug("%s: Unknown EditKey <%s>", File, Key);
	/* Emulate the behavior of "configure" by returning empty string */
	return (char *) NULL;
#ifdef notdef
	/* Return the Key */
	(void) snprintf(Path, sizeof(Path), "@%s@", Key);
#endif
    }

    return (Path[0]) ? Path : (char *) NULL;
}

/*
 * Copy a file and edit the copy.
 * Only works on text files.
 * This is a tad bit slow do to the lack of buffering.
 */
extern int CopyEdit(Src, Dst, Info)
     char		       *Src;
     char		       *Dst;
     McdInfo_t		       *Info;
{
    static char		        Buff[BUFSIZ];
    static char		        Path[MAXPATHLEN];
    char		       *Value;
    FILE		       *SFH;
    FILE		       *DFH;
    int				Status = 0;
    int				LineNo = 1;
    register int		Char;
    register int		i;
    register char	       *cp;

    if (!Src || !Dst)
	return -1;

    Debug("CopyEdit <%s> -> <%s>", Src, Dst);

    SFH = fopen(Src, "r");
    if (!SFH) {
	Error("%s: Open for read failed: %s", Src, SYSERR);
	return -1;
    }

    /*
     * If Dst is a directory then add the base of Src to Dst
     */
    if (IsDir(Dst)) {
	if (cp = strrchr(Src, '/')) {
	    (void) snprintf(Path, sizeof(Path), "%s/%s", Dst, cp+1);;
	    Dst = Path;
	}
    }

    if (MakeFileDir(Dst) < 0) {
	fclose(SFH);
	return -1;
    }

    DFH = fopen(Dst, "w");
    if (!DFH) {
	Error("%s: Open for write/create failed: %s", Dst, SYSERR);
	fclose(SFH);
	return -1;
    }

    /*
     * Read input char by char.  If we find @...@ then lookup ... 
     * to see if it's a known special Key and substitute if it is.
     * If it's not, just print it out.
     */
    while ((Char = fgetc(SFH)) != EOF) {
	if (Char == '\n')
	    ++LineNo;
	if (Char == '@') {
	    /* Fill Buff up with tke Key */
	    for (Char = fgetc(SFH), i = 0; 
		 Char != EOF && Char != '@' && Char != '\n' && i<sizeof(Buff); 
		 Char = fgetc(SFH)) {
		Buff[i++] = Char;
	    }
	    if (i == sizeof(Buff)) {
		Error("%s: line %d: CopyEdit internal Buff to small.", 
		      Dst, LineNo);
		Status = -1;
		break;
	    }

	    /* Terminate the buffer */
	    Buff[i] = CNULL;

	    /* If we hit newline before @, then just print what we have */
	    if (Char == '\n') {
		++LineNo;
		if (fprintf(DFH, "@%s\n", Buff) != i+2) {
		    Error("%s: line %d: write failed: %s", 
			  Dst, LineNo, SYSERR);
		    Status = -1;
		    break;
		}
		continue;
	    }

	    /* Retrieve and write out the Value */
	    Value = EditValue(Buff, Info, Src);
	    if (!Value)
		continue;
	    if (fprintf(DFH, "%s", Value) != strlen(Value)) {
		Error("%s: line %d: write failed: %s", 
		      Dst, LineNo, SYSERR);
		Status = -1;
		break;
	    }

	    if (Char == EOF)
		break;
	    continue;
	}
	if (fputc(Char, DFH) == EOF) {
	    Error("%s: line %d: write failed: %s", Dst, LineNo, SYSERR);
	    Status = -1;
	    break;
	}
    }

    fclose(SFH);
    fclose(DFH);

    return Status;
}
