/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.4 $";
#endif

/*
 * McdInfo functions
 */

#include "defs.h"

/*
 * Remove quotes
 */
static char *DeQuote(String)
     char		       *String;
{
    char		       *New;
    char		       *cp;

    for (cp = String; cp && *cp && (*cp == ' ' || *cp == '"'); ++cp);
    New = cp;

    for (cp = &New[strlen(New)-1]; cp && *cp && (*cp == ' ' || *cp == '"'); 
	 --cp);
    if (cp)
	*(cp+1) = CNULL;

    return New;
}

/*
 * Read mcdinfo file
 */
extern McdInfo_t *McdInfoRead(FileName, DistDir)
     char		       *FileName;
     char		       *DistDir;
{
    FILE		       *FH;
    char			Line[8192];
    char		       *cp;
    int				Argc;
    char		      **Argv;
    int				LineNo = 0;
    int				ErrCount = 0;
    static McdInfo_t		Info;
    char		       *File;

    File = GetMcdFile(FileName, MCDINFO, DistDir);
    if (!File) {
	if (FileName)
	    Error("%s: This %s file doesn't exist.", FileName, MCDINFO);
	else
	    Error("Cannot locate %s in %s.", MCDINFO, DistDir);
	return (McdInfo_t *) NULL;
    }
    McdInfoFile = strdup(File);

    Debug("Reading %s contents from <%s> ...", MCDINFO, File);

    FH = fopen(File, "r");
    if (!FH) {
	Error("%s: Open for read failed: %s", File, SYSERR);
	return (McdInfo_t *) NULL;
    }

    memset(&Info, 0, sizeof(Info));

    /*
     * Read contents
     */
    while (fgets(Line, sizeof(Line), FH)) {
	++LineNo;
	if (Line[0] == '#')
	    continue;
	if (cp = strchr(Line, '\n'))
	    *cp = CNULL;

	Argc = StrToArgv(Line, "=", &Argv, NULL, 0);
#define REQ_FIELDS 2
	if (Argc != REQ_FIELDS) {
	    Error("%s: Line %d: Wrong number of fields detected - found %d require %d",
		  File, LineNo, Argc, REQ_FIELDS);
	    ++ErrCount;
	    continue;
#undef REQ_FIELDS
	}
	if (EQ(Argv[0], "NAME"))	Info.Name = DeQuote(Argv[1]);
	if (EQ(Argv[0], "VERSION"))	Info.Version = DeQuote(Argv[1]);
	if (EQ(Argv[0], "RELEASE"))	Info.Release = DeQuote(Argv[1]);
	if (EQ(Argv[0], "VENDORNAME"))	Info.VendorName = DeQuote(Argv[1]);
	if (EQ(Argv[0], "URL"))	    	Info.URL = DeQuote(Argv[1]);
	if (EQ(Argv[0], "SUMMARY"))	Info.Summary = DeQuote(Argv[1]);
	if (EQ(Argv[0], "OSNAME"))	Info.OSname = DeQuote(Argv[1]);
	if (EQ(Argv[0], "MINOSVERSION")) Info.OSverMin = DeQuote(Argv[1]);
	if (EQ(Argv[0], "OSVERSION"))	Info.OSver = DeQuote(Argv[1]);
	if (EQ(Argv[0], "ARCH"))	Info.Arch = DeQuote(Argv[1]);
	if (EQ(Argv[0], "LIBCNAME"))	Info.LibCname = DeQuote(Argv[1]);
	if (EQ(Argv[0], "INSTALLDIR"))	Info.InstallDir = DeQuote(Argv[1]);
	if (EQ(Argv[0], "INSTALLLIC"))	Info.InstallLic = DeQuote(Argv[1]);
	if (EQ(Argv[0], "MINLIBCVERSION")) 
	    Info.LibCverMin = DeQuote(Argv[1]);
	if (EQ(Argv[0], "ACCEPTLICENSEFILE")) 
	    Info.LicFile = DeQuote(Argv[1]);
	if (EQ(Argv[0], "ASKFORLICENSEFILE") && atoi(DeQuote(Argv[1])))
	    Info.AskForLic = TRUE;
	if (EQ(Argv[0], "DISTBTIME"))	    
	    Info.DistBuilt = strtol(DeQuote(Argv[1]), NULL, 0);
	if (EQ(Argv[0], "SIZE"))	
	    Info.Size = (Large_t) strtol(DeQuote(Argv[1]), NULL, 0);
    }

    fclose(FH);

    if (ErrCount)
	return (McdInfo_t *) NULL;

    return &Info;
}
