/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.19 $";
#endif

/*
 * HTML functions
 */
#include "defs.h"
#include "html.h"
#include "mcsysinfo.h"

static int			LastLineLevel;
static int			OnePassLine;
static int			TableLevel;

/*
 * Get our hostname
 */
static char *HostName()
{
   static MCSIquery_t Query;

   (void) memset(&Query, CNULL, sizeof(Query));
   Query.Cmd = MCSI_HOSTNAME;
   if (mcSysInfo(&Query) == 0)
       return (char *) Query.Out;
   else
       return (char *) NULL;
}

/*
 * Start a table row
 */
extern void HtmlTrStart(Args)
     char		       *Args;
{
    SImsg(SIM_INFO, "\t<TR%s%s>\n",
	  (Args) ? " " : "",
	  (Args) ? Args : "");
}

/*
 * End a table row
 */
extern void HtmlTrEnd()
{
    SImsg(SIM_INFO, "\t</TR>\n");
}

/*
 * Start a table TD
 */
extern void HtmlTdStart(Args)
     char		       *Args;
{
    SImsg(SIM_INFO, "\t<TD%s%s>\n",
	  (Args) ? " " : "",
	  (Args) ? Args : "");
}

/*
 * End a table TD
 */
extern void HtmlTdEnd()
{
    SImsg(SIM_INFO, "\t</TD>\n");
}

/*
 * Display HTML header
 */
extern void HtmlStart()
{
    static char			Buff[512];
    char		       *Host = HostName();

    SImsg(SIM_INFO, 
        "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n");
    SImsg(SIM_INFO, "<HTML>\n");
    SImsg(SIM_INFO, "\t<TITLE>SysInfo for %s</TITLE>\n", PRTS(Host));

    SImsg(SIM_INFO, "<BODY bgcolor=\"#ffffff\">\n");
    SImsg(SIM_INFO, "<CENTER>\n");
    SImsg(SIM_INFO, "<H1>\n");
    SImsg(SIM_INFO,
	  "<a href=\"http://www.MagniComp.com/sysinfo/\">SysInfo</a> for %s\n",
	  PRTS(Host));
    SImsg(SIM_INFO, "</H1>\n");
    SImsg(SIM_INFO, "</CENTER>\n");
}

/*
 * End HTML body
 */
extern void HtmlEnd()
{
    time_t			Now;

    SImsg(SIM_INFO, "\n<HR>\n");
    SImsg(SIM_INFO, "<TABLE border=0 cellpadding=0 cellspacing=0>");
    SImsg(SIM_INFO, "<TR>\n");

    SImsg(SIM_INFO, "<TD align=left>\n");
    SImsg(SIM_INFO, "<a href=\"http://www.magnicomp.com/\">");
    SImsg(SIM_INFO, "<IMG src=\"http://www.MagniComp.com/images/magnicomp-logo-2.gif\" alt=\"www.MagniComp.com\" border=0>\n");
    SImsg(SIM_INFO, "</a>\n");
    SImsg(SIM_INFO, "</TD>\n");

    SImsg(SIM_INFO, "<TD NOWRAP align=left>\n");
    SImsg(SIM_INFO, "<FONT size=-1>\n");
    SImsg(SIM_INFO, "Autogenerated by: ");
    SImsg(SIM_INFO, 
	  "<a href=\"http://www.magnicomp.com/\">MagniComp's</a> ");
    SImsg(SIM_INFO, 
	  "<a href=\"http://www.magnicomp.com/sysinfo/\">SysInfo</a><BR>\n");
    SImsg(SIM_INFO, "Generated on: %s<BR>\n", HostName());
    time(&Now);
    SImsg(SIM_INFO, "Generated: %s<BR>\n", TimeStr(Now, "%c"));
    SImsg(SIM_INFO, "</FONT>\n");
    SImsg(SIM_INFO, "</TD>\n");
    SImsg(SIM_INFO, "</TR>\n");

    SImsg(SIM_INFO, "</TABLE>\n");
    SImsg(SIM_INFO, "</BODY>\n");
    SImsg(SIM_INFO, "</HTML>\n");
}

/*
 * New line
 */
extern void HtmlNewLine()
{
    SImsg(SIM_INFO, "<BR>\n");
}

/*
 * Display 1 line possibly offset by OffSet
 * A side effect is that all new text is also indented by OffSet until
 * HtmlLineEnd() is called.
 */
extern void HtmlLine(OffSet, String)
     int			OffSet;
     char		       *String;
{
    int				Level;
    int				i;

    Level = OffSet / OffSetAmt;

    if (OffSet >= 0) {
	/* 
	 * We always start with UL on first pass so we get no "circle"
	 * marks at the start of the line.
	 */
	if (!OnePassLine && OffSet == 0) {
	    SImsg(SIM_INFO, "<UL type=\"none\">\n");
	} else if (Level > LastLineLevel) {
	    /* Increase indent until we reach Level */
	    for (i = LastLineLevel; i < Level; ++i)
		SImsg(SIM_INFO, "<UL type=\"none\">\n");
	} else if (Level < LastLineLevel) {
	    /* Decrease indent from LastLineLevel back down to Level */
	    for (i = LastLineLevel; i > Level; --i)
		SImsg(SIM_INFO, "</UL>\n");
	}

	SImsg(SIM_INFO, "<LI>%s<BR>\n", PRTS(String));

	OnePassLine = 1;
    } else {
	if (String)
	    SImsg(SIM_INFO, "%s", String);
    }

    if (OffSet >= 0)
	LastLineLevel = Level;
}

/*
 * End any previously started HtmlLine()'s
 */
extern void HtmlLineEnd()
{
    register int		i;

    if (LastLineLevel)
	for (i = 0; i < LastLineLevel; ++i)
	    SImsg(SIM_INFO, "</UL>\n");
    else if (OnePassLine)
	SImsg(SIM_INFO, "</UL>\n");

    LastLineLevel = -1;
}

/*
 * Display String as a banner
 */
extern void HtmlBanner(String)
     char		       *String;
{
    SImsg(SIM_INFO, "<CENTER><H1>%s</H1></CENTER>\n", String);
}

/*
 * Start HTML Table
 */
extern void HtmlTableStart(Table)
     HtmlTable_t	       *Table;
{
    static char		        Args[512];
    char			Buff[256];

    Args[0] = CNULL;
    if (Table) {
	snprintf(Args, sizeof(Args), 
		 " border=%d cellpadding=%d cellspacing=%d",
		 Table->Border,
		 Table->CellPadding,
		 Table->CellSpacing);
	if (Table->Style) {
	    snprintf(Buff, sizeof(Buff), " style=\"%s\"",
		     Table->Style);
	    StrAppend(Args, sizeof(Args), Buff);
	}
	if (Table->Width) {
	    snprintf(Buff, sizeof(Buff), " width=%d",
		     Table->Width);
	    StrAppend(Args, sizeof(Args), Buff);
	}
	if (Table->Height) {
	    snprintf(Buff, sizeof(Buff), " height=%d",
		     Table->Height);
	    StrAppend(Args, sizeof(Args), Buff);
	}
    }

    /* Track our depth level */
    ++TableLevel;

    /*
     * We enclose the table inside another table to allow us to properly
     * align and center the title.
     */
    SImsg(SIM_INFO, "\n<!-- HtmlTableStart(%d) -->\n", TableLevel);
    SImsg(SIM_INFO, "<TABLE border=0 cellpadding=0 cellspacing=0>\n");
    SImsg(SIM_INFO, "<TR><TD><TABLE%s>\n", PRTS(Args));
	  
    if (Table && Table->Title)
	/*
	 * It would be best if we could set the bgcolor for the CAPTION
	 * but we can't do that with HTML at the moment so we'll set the
	 * font color.
	 */
	SImsg(SIM_INFO, 
	      "<CAPTION align=top><font color=\"%s\"><b>%s</b></font></CAPTION>\n",
	      HTML_BGHEAD_1, Table->Title);
}

/*
 * End HTML Table
 * XXX Assumes HtmlTableStart() was used to start the table.
 */
extern void HtmlTableEnd()
{
    if (TableLevel > 0) {
	SImsg(SIM_INFO, "\n</TD></TR></TABLE>\n");
	SImsg(SIM_INFO, "</TABLE>\n");
	SImsg(SIM_INFO, "<!-- HtmlTableEnd(%d) -->\n", TableLevel);
	--TableLevel;
    }
}

/*
 * Table Header Line
 */
extern void HtmlTableHeader(Argv, Argc, FormatArgv, Args)
     char		      **Argv;
     int			Argc;
     char		      **FormatArgv;
     char		       *Args;
{
    register int		i;

    if (!Argv)
	return;

    SImsg(SIM_INFO, "<TR %s>\n", PRTS(Args));
    for (i = 0; (Argc && i < Argc) || (!Argc && Argv[i]); ++i) {
	SImsg(SIM_INFO, "\t<TH %s>%s</TH>\n",
	      (FormatArgv && FormatArgv[i] && FormatArgv[i][0]) ?
	      FormatArgv[i] : "",
	      (Argv[i] && Argv[i][0]) ? Argv[i] :
	      "&nbsp;");
    }
    SImsg(SIM_INFO, "</TR>\n");
}

/*
 * Table data line
 */
extern void HtmlTableData(DataArgv, DataArgc, FormatArgv, RowArgs)
     char		      **DataArgv;
     int			DataArgc;
     char		      **FormatArgv;
     char		       *RowArgs;
{
    register int		i;

    SImsg(SIM_INFO, "<TR %s>\n", PRTS(RowArgs));
    for (i = 0; i < DataArgc; ++i) {
	SImsg(SIM_INFO, "\t<TD %s>%s</TD>\n",
	      (FormatArgv && FormatArgv[i] && FormatArgv[i][0]) ?
	      FormatArgv[i] : "",
	      (DataArgv[i] && DataArgv[i][0]) ?
	      DataArgv[i] : "&nbsp;");
    }
    SImsg(SIM_INFO, "</TR>\n");
}

/*
 * Start of a ShowLabel block
 */
extern void HtmlShowLabelStart()
{
    static HtmlTable_t	        Table;

    Table.Style = "margin-left: .5in";
    /* We want to keep things looking together and tight */
    Table.CellSpacing = 0;
    Table.CellPadding = 1;
    HtmlTableStart(&Table);
}

/*
 * End of a ShowLabel block
 */
extern void HtmlShowLabelEnd()
{
    HtmlTableEnd();
}

extern void HtmlShowLabel(Name, Value, Level)
     char		       *Name;
     char		       *Value;
     int			Level;
{
    static int			Toggle;
    char		       *Color;

    if (!Toggle) {
	Color = HTML_BGLINE_2A;
	Toggle = 1;
    } else {
	Color = HTML_BGLINE_2B;
	Toggle = 0;
    }

    SImsg(SIM_INFO, "<TR bgcolor=\"%s\">\n", Color);
    SImsg(SIM_INFO, "\t<TD NOWRAP><b><FONT size=-1>%s:</FONT></b></TD>\n",
	  (Name) ? Name : "&nbsp;");
    SImsg(SIM_INFO, "\t<TD>&nbsp;</TD>\n");	/* Spacer */
    SImsg(SIM_INFO, "\t<TD><FONT size=-1><kbd>%s</kbd></FONT></TD>\n",
	  (Value && Value[0]) ? Value : "&nbsp;");
    SImsg(SIM_INFO, "</TR>\n");
}

/*
 * Start or end an HtmlKeyValue() sequence.
 */
extern void HtmlKeyValueSet()
{
    static int			Show;

    if (!Show) {
	SImsg(SIM_INFO, "\n<Table BORDER=0 CELLPADDING=0 CELLSPACING=0>\n");
	Show = 1;
    } else {
	SImsg(SIM_INFO, "\n</Table>\n");
	Show = 0;
    }
}

/*
 * Display a Key/Value pair
 */
extern void HtmlKeyValue(Label, Key, Value)
     char		       *Label;
     char		       *Key;
     char		       *Value;
{
    static int			Toggle;
    char		       *Color;

    if (!Value || !*Value)
	return;

    if (!Toggle) {
	Color = HTML_BGLINE_1A;
	Toggle = 1;
    } else {
	Color = HTML_BGLINE_1B;
	Toggle = 0;
    }

    SImsg(SIM_INFO, "<TR VALIGN=TOP BGCOLOR=\"%s\">\n", Color);
    if (VL_TERSE) {
	SImsg(SIM_INFO, "\t<TD><kbd>%s</kbd></TD>\n", Value);
    } else if (VL_BRIEF) {
	SImsg(SIM_INFO, "\t<TD NOWRAP><b>%s</b></TD>\n", Key);
	SImsg(SIM_INFO, "\t<TD></TD><TD>&nbsp;</TD>\n");	/* Spacer */
	SImsg(SIM_INFO, "\t<TD valign=center><kbd>%s</kbd></TD>\n", Value);
    } else {
	if (Label) {
	    SImsg(SIM_INFO, "\t<TD NOWRAP><B>%s</B>\n", Label);
	    if (VL_ALL)
		SImsg(SIM_INFO, " (<i>%s</i>)", Key);
	} else
	    SImsg(SIM_INFO, "<TD>%s", Key);
	SImsg(SIM_INFO, "</TD>\n");
	SImsg(SIM_INFO, "\t<TD></TD><TD>&nbsp;</TD>\n");	/* Spacer */
	SImsg(SIM_INFO, "\t<TD valign=center><kbd>%s</kbd></TD>\n", Value);
    }
    SImsg(SIM_INFO, "</TR>\n");
}

/*
 * Bold Name
 */
extern char *HtmlBold(Name)
     char		       *Name;
{
    static char			Buff[256];

    snprintf(Buff, sizeof(Buff), "<b>%s</b>", Name);

    return Buff;
}

/*
 * Italic Name
 */
extern char *HtmlItalic(Name)
     char		       *Name;
{
    static char			Buff[256];

    snprintf(Buff, sizeof(Buff), "<i>%s</i>", Name);

    return Buff;
}

/*
 * Encode Name
 */
extern char *HtmlEncodeName(Name)
     char		       *Name;
{
    return HtmlBold(Name);
}

