/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.25 $";
#endif

/*
 * Show Partition Info related functions.
 */

#include "defs.h"
#include "html.h"

extern int			DoPrintUnused;

/*
 * List valid arguments for Partition class. 
 */
extern void ListPartition()
{
    SImsg(SIM_INFO, 
	  "\n\nTo show specific partition use `-class partition -show part1,part2...'.\n\n");
}

/*
 * Display details of a PartInfo entry as OE_TEXT and OE_HTML
 */
static void ShowPartInfoDetailsText(Part, OffSet)
     PartInfo_t		       *Part;
     int			OffSet;
{
    char		      **cpp;
    static char			Buff[256];

    if (!VL_DESC && !VL_CONFIG)
	return;

#define CHK(a)	( a && *(a) )
    ShowLabelStart();
    
    ShowLabel("Name", Part->BaseName, OffSet);
    
    if (CHK(Part->DevPath))
	ShowLabel("Device Path", Part->DevPath, OffSet);
    if (CHK(Part->DevPathRaw))
	ShowLabel("Device Path Raw", Part->DevPathRaw, OffSet);
    if (Part->UsageStatus)
	ShowLabel("Usage Status", Part->UsageStatus, OffSet);
    if (CHK(Part->MntName))
	ShowLabel("Mount/Usage", Part->MntName, OffSet);
    if (CHK(Part->MntOpts)) {
	Buff[0] = CNULL;
	for (cpp = Part->MntOpts; cpp && *cpp; ++cpp) {
	    if (Buff[0])
		StrAppend(Buff, sizeof(Buff), " ");
	    StrAppend(Buff, sizeof(Buff), *cpp);
	}
	ShowLabel("Mount/Usage Options", Buff, OffSet);
    }
    if (CHK(Part->Type))
	ShowLabel("Type", Part->Type, OffSet);
    if (CHK(Part->TypeDesc))
	ShowLabel("Type Description", Part->TypeDesc, OffSet);
    if (Part->TypeNum) {
	snprintf(Buff, sizeof(Buff), "0x%02X", Part->TypeNum);
	ShowLabel("Type Number", Buff, OffSet);
    }
    if (Part->Size)
	ShowLabel("Size", GetSizeStr(Part->Size, MBYTES), OffSet);
    if (Part->AmtUsed)
	ShowLabel("Amount In Use", GetSizeStr(Part->AmtUsed, MBYTES), 
		  OffSet);
    if (Part->SecSize)
	ShowLabel("Sector Size", itoa(Part->SecSize), OffSet);
    /* Assume StartSect is valid only if NumSect is */
    if (Part->StartSect && Part->NumSect)
	ShowLabel("Starting Sector", Ltoa(Part->StartSect), OffSet);
    if (Part->EndSect && Part->StartSect)
	ShowLabel("Ending Sector", Ltoa(Part->EndSect), OffSet);
    if (Part->NumSect)
	ShowLabel("# of Sectors", Ltoa(Part->NumSect), OffSet);

    ShowLabelEnd();
#undef CHK
}

/*
 * Switch for Details of PartInfo
 */
static void ShowPartInfoDetails(Part, OffSet)
     PartInfo_t		       *Part;
     int			OffSet;
{
    switch (EncodeType) {
    case OE_HTML:
    case OE_TEXT:
	ShowPartInfoDetailsText(Part, OffSet);
	break;
    }
}

/*
 * Tree Display of a PartInfo entry.
 */
static void ShowPartInfoTree(Part, OffSet)
     PartInfo_t		       *Part;
     int			OffSet;
{
    char		      **cpp;
    static char			Buff[256];

    if (!Part)
	return;

    snprintf(Buff, sizeof(Buff), "%-15s", ShowEncodeName(Part->BaseName));

    if (Part->MntName || Part->Type || Part->Size)
	StrAppend(Buff, sizeof(Buff), " is a ");

    if (Part->Size) {
	StrAppend(Buff, sizeof(Buff), GetSizeStr(Part->Size, MBYTES));
	StrAppend(Buff, sizeof(Buff), " ");
    }

    if (Part->TypeDesc) {
	StrAppend(Buff, sizeof(Buff), Part->TypeDesc);
	StrAppend(Buff, sizeof(Buff), " ");
    } else if (Part->Type) {
	StrAppend(Buff, sizeof(Buff), Part->Type);
	StrAppend(Buff, sizeof(Buff), " ");
    } else if (Part->Usage == PIU_UNUSED) {
	StrAppend(Buff, sizeof(Buff), "UNUSED ");
    }

    StrAppend(Buff, sizeof(Buff), "partition ");

    if (Part->MntName) {
	StrAppend(Buff, sizeof(Buff), "in use as ");
	StrAppend(Buff, sizeof(Buff), Part->MntName);
    }

    ShowLine(OffSet, Buff);

    ShowPartInfoDetails(Part, OffSet);
}

/*
 * Report format Display of a PartInfo entry.
 */
static void ShowPartInfoReport(Part, OffSet)
     PartInfo_t		       *Part;
     int			OffSet;
{
    static char		       *RptData[17];
    char		       *cp;

    (void) memset(RptData, CNULL, sizeof(RptData));
    RptData[0] = PRTS(Part->DevName);
    RptData[1] = PRTS(Part->BaseName);
    RptData[2] = PRTS(Part->Name);
    RptData[3] = strdup(itoa(Part->Num));
    RptData[4] = PRTS(Part->DevPathRaw);
    RptData[5] = PRTS(Part->Type);
    RptData[6] = PRTS(Part->TypeDesc);
    RptData[7] = strdup(itoa(Part->TypeNum));
    RptData[8] = PRTS(Part->UsageStatus);
    RptData[9] = PRTS(Part->MntName);
    if (ArgvToStr(&cp, Part->MntOpts, 0, ",", 0) == 0)
	RptData[10] = PRTS(cp);
    RptData[11] = strdup(Ltoa(Part->Size));
    RptData[12] = strdup(Ltoa(Part->AmtUsed));
    RptData[13] = strdup(itoa(Part->SecSize));
    RptData[14] = strdup(Ltoa(Part->StartSect));
    RptData[15] = strdup(Ltoa(Part->EndSect));
    RptData[16] = strdup(Ltoa(Part->NumSect));

    Report(CN_PARTITION, R_NAME, PRTS(Part->DevPath), RptData,
	   sizeof(RptData) / sizeof(char *));
}

/*
 * Column Display of a PartInfo entry as OE_TEXT
 */
extern void ShowPartInfoColumnsText(Part, OffSet)
     PartInfo_t		       *Part;
     int			OffSet;
{
    char		      **cpp;
    static char			Cap[256];
    static char			AmtUsed[256];
    static char		       *FormatStr;
    char		      **Format;
    ClassInfo_t		       *Class;

    if (!Part)
	return;

    if (!FormatStr) {
	Class = ClassGetByName(CN_PARTITION);
	if (!Class) {
	    SImsg(SIM_DBG, "ShowPartInfoColumnsText: Cannot find %s class",
		  CN_PARTITION);
	    return;
	}
	Format = ColGetFmt(Class->Headers);
	if (!Format)
	    return;

	FormatStr = ColGetFmtStr(Format);
	if (!FormatStr)
	    return;
    }

    snprintf(Cap, sizeof(Cap), "%s",
	     (Part->Size) ? GetSizeStr(Part->Size, MBYTES) : "-");
    snprintf(AmtUsed, sizeof(AmtUsed), "%s",
	     (Part->AmtUsed) ? GetSizeStr(Part->AmtUsed, MBYTES) : "-");

    SImsg(SIM_INFO, FormatStr,
	  CPRT(Part->BaseName),
	  Cap,
	  AmtUsed,
	  CPRT(Part->Type),
	  CPRT(Part->MntName));
}

/*
 * Column Display of a PartInfo entry as OE_HTML
 */
extern void ShowPartInfoColumnsHtml(Part, OffSet)
     PartInfo_t		       *Part;
     int			OffSet;
{
    static char			Cap[256];
    static char			AmtUsed[256];
    char		       *DataArgv[5];
    static int			Counter;
    char		        Options[128];

    if (!Part)
	return;

    snprintf(Cap, sizeof(Cap), "%s",
	     (Part->Size) ? GetSizeStr(Part->Size, MBYTES) : HTML_SPACE);
    snprintf(AmtUsed, sizeof(AmtUsed), "%s",
	     (Part->AmtUsed) ? GetSizeStr(Part->AmtUsed, MBYTES) : HTML_SPACE);

    DataArgv[0] = PRTS(Part->BaseName);
    DataArgv[1] = Cap;
    DataArgv[2] = AmtUsed;
    DataArgv[3] = PRTS(Part->Type);
    DataArgv[4] = PRTS(Part->MntName);

    /* Rows alternate colors */
    if (Counter++ % 2)
	snprintf(Options, sizeof(Options), "bgcolor=\"%s\"",
		 HTML_BGLINE_1A);
    else
	snprintf(Options, sizeof(Options), "bgcolor=\"%s\"",
		 HTML_BGLINE_1B);

    HtmlTableData(DataArgv, sizeof(DataArgv)/sizeof(char *), NULL, Options);
}

/*
 * Display Partition info
 */
extern int ShowPartition(MyClass, Names)
    ClassInfo_t		       *MyClass;
    char		      **Names;
{
    static PartInfo_t 	       *Root = NULL;
    static MCSIquery_t		Query;
    PartInfo_t		       *Part;

    (void) memset(&Query, CNULL, sizeof(Query));
    Query.Op = MCSIOP_CREATE;
    Query.Cmd = MCSI_PARTITION;
    if (mcSysInfo(&Query) != 0)
	return CLASS_UNSUPPORTED;

    Root = (PartInfo_t *) Query.Out;

    if (!Root || !Root->Next)
	return CLASS_UNSUPPORTED;

    ShowStart(MyClass, (int *)NULL);

    /*
     * Traverse list
     */
    for (Part = Root; Part; Part = Part->Next) {
	if (!Part->BaseName || (Part->Type == PIU_UNUSED && !DoPrintUnused))
	    continue;
	switch (FormatType) {
	case FT_TREE:
	    ShowPartInfoTree(Part, 0);
	    break;
	case FT_COLUMNS:
	    switch (EncodeType) {
	    case OE_TEXT:
		ShowPartInfoColumnsText(Part, 0);
		ShowPartInfoDetails(Part, 0);
		break;
	    case OE_HTML:
		ShowPartInfoColumnsHtml(Part, 0);
		ShowPartInfoDetails(Part, 0);
		break;
	    }
	    break;
	case FT_REPORT:
	    ShowPartInfoReport(Part, 0);
	    break;
	}
    }

    ShowEnd(MyClass);

    return 0;
}
