/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.35 $";
#endif

/*
 * Show Software Info related functions.
 */

#include "defs.h"
#include "html.h"

static void 			DisplaySoftInfoNode();
extern int			SwFiles;

/*
 * List valid arguments for Software class. 
 */
extern void ListSoftInfo()
{
    SImsg(SIM_INFO, 
"\n\nTo show specific software use `-class software -show pkg1,pkg2...'.\n\n");
}

/*
 * Display Software Info
 */
extern int ShowSoftInfo(MyClass, Names)
    ClassInfo_t		       *MyClass;
    char		      **Names;
{
    static SoftInfo_t 	       *Root = NULL;
    static MCSIquery_t		Query;

    (void) memset(&Query, CNULL, sizeof(Query));
    Query.Op = MCSIOP_CREATE;
    Query.Cmd = MCSI_SOFTINFO;
    if (mcSysInfo(&Query) != 0) {
	return -1;
    }
    Root = (SoftInfo_t *) Query.Out;

    if (!Root || !(Root->Slaves || Root->Next))
	return CLASS_UNSUPPORTED;

    ShowStart(MyClass, 0);

    DisplaySoftInfoNode(Root, 0);

    ShowEnd(MyClass);

    return 0;
}

/*
 * Display details of SoftInfo as OE_TEXT
 */
static void DetailsSoftInfoText(SoftInfo, OffSet)
     SoftInfo_t		       *SoftInfo;
     int			OffSet;
{
    register SoftInfo_t	       *sp;
    register SoftFileList_t    *SoftFile;
    Desc_t		       *Desc;
    static char		       *Argv[6];
    static HtmlTable_t		Table;
    static char		       *FormatArgv[6];
    static char			Buff[64 * KBYTES];
    int				Count = 0;

    if (!VL_DESC && !VL_CONFIG)
	return;

#define CHK(a)	( a && *(a) )

    ShowLabelStart();
    
    ShowLabel("Name", SoftInfo->Name, OffSet);
    
    if (CHK(SoftInfo->Version))
	ShowLabel("Version", SoftInfo->Version, OffSet);
    if (CHK(SoftInfo->Revision))
	ShowLabel("Revision", SoftInfo->Revision, OffSet);
    if (CHK(SoftInfo->Category))
	ShowLabel("Category", SoftInfo->Category, OffSet);
    if (CHK(SoftInfo->SubCategory))
	ShowLabel("Sub Category", SoftInfo->SubCategory, OffSet);
    if (CHK(SoftInfo->OSname))
	ShowLabel("OS Name", SoftInfo->OSname, OffSet);
    if (CHK(SoftInfo->OSversion))
	ShowLabel("OS Version", SoftInfo->OSversion, OffSet);
    if (CHK(SoftInfo->Arch))
	ShowLabel("Architecture", SoftInfo->Arch, OffSet);
    if (CHK(SoftInfo->ISArch))
	ShowLabel("Instruct Set Arch", SoftInfo->ISArch, OffSet);
    if (CHK(SoftInfo->BuildDate))
	ShowLabel("Date Built", SoftInfo->BuildDate, OffSet);
    if (CHK(SoftInfo->InstDate))
	ShowLabel("Date Installed", SoftInfo->InstDate, OffSet);
    if (CHK(SoftInfo->ProdStamp))
	ShowLabel("Production Stamp", SoftInfo->ProdStamp, OffSet);
    if (CHK(SoftInfo->BaseDir))
	ShowLabel("Base Directory", SoftInfo->BaseDir, OffSet);
    if (CHK(SoftInfo->VendorName))
	ShowLabel("Vendor Name", SoftInfo->VendorName, OffSet);
    if (CHK(SoftInfo->VendorEmail))
	ShowLabel("Vendor Email", SoftInfo->VendorEmail, OffSet);
    if (CHK(SoftInfo->VendorPhone))
	ShowLabel("Vendor Phone", SoftInfo->VendorPhone, OffSet);
    if (CHK(SoftInfo->VendorStock))
	ShowLabel("Vendor Stock #", SoftInfo->VendorStock, OffSet);
    
    for (Desc = SoftInfo->DescList; Desc; Desc = Desc->Next) {
	ShowLabel((Desc->Label) ? Desc->Label : "Description", 
		  Desc->Desc, OffSet);
    }
    
    if (SoftInfo->Master && SoftInfo->Master->Name) {
	snprintf(Buff, sizeof(Buff), "%s", SoftInfo->Master->Name);
	if (SoftInfo->Master->Version) {
	    StrAppend(Buff, sizeof(Buff), " ");
	    StrAppend(Buff, sizeof(Buff), SoftInfo->Master->Version);
	}
	ShowLabel("Part of Product", Buff, OffSet);
    }
    if (SoftInfo->Slaves) {
	Buff[0] = CNULL;
	for (sp = SoftInfo->Slaves; sp; sp = sp->Next) {
	    if (Buff[0])
		StrAppend(Buff, sizeof(Buff), " ");
	    StrAppend(Buff, sizeof(Buff), sp->Name);
	}
	ShowLabel("Product Components", Buff, OffSet);
    }
    if (SoftInfo->DiskUsage)
	ShowLabel("Disk Usage", GetSizeStr(SoftInfo->DiskUsage, 1),OffSet);
    if (CHK(SoftInfo->License))
	ShowLabel("License", SoftInfo->License, OffSet);
    if (CHK(SoftInfo->Copyright))
	ShowLabel("Copyright", SoftInfo->Copyright, OffSet);
    if (CHK(SoftInfo->URL))
	ShowLabel("Product URL", SoftInfo->URL, OffSet);
    if (CHK(SoftInfo->Desc))
	ShowLabel("Description", SoftInfo->Desc, OffSet);
    
    if (CHK(SoftInfo->DescVerbose)) {
	if (FormatType == FT_TREE && strchr(SoftInfo->DescVerbose, '\n'))
	    snprintf(Buff, sizeof(Buff), "\\\n%s",
		     SoftInfo->DescVerbose);
	else
	    snprintf(Buff, sizeof(Buff), "%s", SoftInfo->DescVerbose);
	ShowLabel("Verbose Description", Buff, OffSet);
    }
    
    if (SoftInfo->FileList) {
	if (SwFiles) {
	    ShowLabel("Files", NULL, OffSet);
	    switch (FormatType) {
	    case FT_TREE:
		switch (EncodeType) {
		case OE_TEXT:
		    for (SoftFile = SoftInfo->FileList; SoftFile; 
			 SoftFile = SoftFile->Next) {
			if (!SoftFile->FileData)
			    continue;
			ShowOffSet(OffSet + 8);
			SImsg(SIM_INFO, "%s", SoftFile->FileData->Path);
			if (SoftFile->FileData->LinkTo)
			    SImsg(SIM_INFO, " -> %s", 
				  SoftFile->FileData->LinkTo);
			if (SoftFile->FileData->FileSize)
			    SImsg(SIM_INFO, "  SIZE: %s",
				  GetSizeStr(SoftFile->FileData->FileSize, 1));
			if (SoftFile->FileData->MD5)
			    SImsg(SIM_INFO, ", MD5: %s", SoftFile->FileData->MD5);
			if (SoftFile->FileData->CheckSum)
			    SImsg(SIM_INFO, ", CHECKSUM: %s", 
				  SoftFile->FileData->CheckSum);
			SImsg(SIM_INFO, "\n");
		    }
		    break;
		case OE_HTML:
		    Table.Style = "margin-left: .5in";
		    HtmlTableStart(&Table);
		    
		    Argv[0] = "Path";
		    Argv[1] = "Link To";
		    Argv[2] = "Size";
		    Argv[3] = "MD5";
		    Argv[4] = "Checksum";
		    FormatArgv[0] = "valign=bottom";
		    FormatArgv[1] = "valign=bottom";
		    FormatArgv[2] = "valign=bottom align=right";
		    FormatArgv[3] = "valign=bottom align=left";
		    FormatArgv[4] = "valign=bottom align=left";
		    snprintf(Buff, sizeof(Buff), "bgcolor=\"%s\"",
			     HTML_BGHEAD_2);
		    HtmlTableHeader(Argv, 5, FormatArgv, Buff);
		    
		    FormatArgv[0] = "NOWRAP";
		    FormatArgv[1] = "NOWRAP";
		    FormatArgv[2] = "NOWRAP align=right";
		    FormatArgv[3] = "NOWRAP align=left";
		    FormatArgv[4] = "NOWRAP align=left";
		    
		    for (SoftFile = SoftInfo->FileList; SoftFile; 
			 SoftFile = SoftFile->Next) {
			if (!SoftFile->FileData)
			    continue;
			Argv[0] = SoftFile->FileData->Path;
			Argv[1] = SoftFile->FileData->LinkTo;
			Argv[2] = (SoftFile->FileData->FileSize) ?
			    GetSizeStr(SoftFile->FileData->FileSize, 1) : "";
			Argv[3] = SoftFile->FileData->MD5;
			Argv[4] = SoftFile->FileData->CheckSum;
			snprintf(Buff, sizeof(Buff), "bgcolor=\"%s\"",
				 (Count++ % 2) ? HTML_BGLINE_2A : 
				 HTML_BGLINE_2B);
			HtmlTableData(Argv, 5, FormatArgv, Buff);
		    }
		    HtmlTableEnd();
		    break;
		}
		break;
	    }
	} else {
	    ShowLabel("Files", "Specify `+swfiles' to see list of files",
		      OffSet);
	}
    }

    ShowLabelEnd();
#undef CHK
}

/*
 * Switch for DetailsSoftInfo*
 */
static void DetailsSoftInfo(SoftInfo, OffSet)
     SoftInfo_t		       *SoftInfo;
     int			OffSet;
{
    switch (EncodeType) {
    case OE_HTML:
    case OE_TEXT:
	DetailsSoftInfoText(SoftInfo, OffSet);
	break;
    }
}

/*
 * Column Display of a SoftInfo entry as FT_COLUMNS OE_TEXT
 */
static void DisplaySoftInfoColumnsText(SoftInfo, OffSet)
     SoftInfo_t		       *SoftInfo;
     int			OffSet;
{
    static char			Buff[128];
    ClassInfo_t		       *Class;
    static char		       *FormatStr;
    char		      **Format;
    char		       *cp;

    if (!SoftInfo->Name)
	return;

    /*
     * If this is a Product, then handle specially
     */
    if (SoftInfo->EntryTypeNum == MC_SET_PROD) {
	snprintf(Buff, sizeof(Buff), "\t*** %s", SoftInfo->Name);
	if (SoftInfo->Version) {
	    StrAppend(Buff, sizeof(Buff), " ");
	    StrAppend(Buff, sizeof(Buff), SoftInfo->Version);
	}
	StrAppend(Buff, sizeof(Buff), " ***");
	SImsg(SIM_INFO, "%s\n", Buff);
	return;
    }

    if (!FormatStr) {
	Class = ClassGetByName(CN_SOFTINFO);
	if (!Class) {
	    SImsg(SIM_DBG, "DisplaySoftInfoColumns: Cannot find %s class",
		  CN_SOFTINFO);
	    return;
	}
	Format = ColGetFmt(Class->Headers);
	if (!Format)
	    return;

	FormatStr = ColGetFmtStr(Format);
	if (!FormatStr)
	    return;
    }

    snprintf(Buff, sizeof(Buff), "%s", SoftInfo->Desc);
    if (cp = strchr(Buff, '\n'))
	*cp = CNULL;

    SImsg(SIM_INFO, FormatStr,
	  CPRT(SoftInfo->Name),
	  CPRT(SoftInfo->Version),
	  Buff);

    DetailsSoftInfo(SoftInfo, OffSet);
}

/*
 * Column Display of a SoftInfo entry as FT_COLUMNS OE_HTML
 */
static void DisplaySoftInfoColumnsHtml(SoftInfo, OffSet)
     SoftInfo_t		       *SoftInfo;
     int			OffSet;
{
    static char			Buff[256];
    char		       *cp;
    char		       *DataArgv[3];
    int				DataArgc = 3;
    char		       *FormArgv[3];
    static int			Counter;
    char		        Options[128];

    if (!SoftInfo->Name)
	return;

    DataArgv[0] = DataArgv[1] = DataArgv[2] = NULL;
    FormArgv[0] = FormArgv[1] = FormArgv[2] = NULL;

    if (SoftInfo->EntryTypeNum == MC_SET_PROD) {
	/*
	 * Product
	 */
	snprintf(Buff, sizeof(Buff), "%s ", SoftInfo->Name);
	if (SoftInfo->Version)
	    StrAppend(Buff, sizeof(Buff), SoftInfo->Version);
	DataArgc = 1;
	DataArgv[0] = HtmlItalic(Buff);
	FormArgv[0] = "colspan=3 align=center";
	
	snprintf(Options, sizeof(Options), "bgcolor=\"%s\"",
		 HTML_BGHEAD_2);
    } else {
	/*
	 * Package
	 */
	Buff[0] = CNULL;
	if (SoftInfo->Desc) {
	    snprintf(Buff, sizeof(Buff), "%s", SoftInfo->Desc);
	    if (cp = strchr(Buff, '\n'))
		*cp = CNULL;
	}

	DataArgv[0] = PRTS(SoftInfo->Name);
	DataArgv[1] = PRTS(SoftInfo->Version);
	DataArgv[2] = PRTS(Buff);

	/* Rows alternate colors */
	if (Counter++ % 2)
	    snprintf(Options, sizeof(Options), "bgcolor=\"%s\"",
		     HTML_BGLINE_1A);
	else
	    snprintf(Options, sizeof(Options), "bgcolor=\"%s\"",
		     HTML_BGLINE_1B);
    }

    HtmlTableData(DataArgv, DataArgc, FormArgv, Options);

    DetailsSoftInfo(SoftInfo, OffSet);
}

/*
 * Column Display of a SoftInfo entry as FT_COLUMNS
 */
static void DisplaySoftInfoColumns(SoftInfo, OffSet)
     SoftInfo_t		       *SoftInfo;
     int			OffSet;
{
    switch (EncodeType) {
    case OE_TEXT:
	DisplaySoftInfoColumnsText(SoftInfo, OffSet);
	break;
    case OE_HTML:
	DisplaySoftInfoColumnsHtml(SoftInfo, OffSet);
	break;
    }
}

/*
 * Tree Display of a SoftInfo entry.
 */
static void DisplaySoftInfoTree(SoftInfo, OffSet)
     SoftInfo_t		       *SoftInfo;
     int			OffSet;
{
    Desc_t		       *Desc;
    char		       *cp;
    static char			Buff[64 * KBYTES];
    static char			Buff2[512];

    if (!SoftInfo->Name)
	return;

    if (FormatType == FT_TREE)
	snprintf(Buff, sizeof(Buff), "%-15s   ",
		 ShowEncodeName(SoftInfo->Name));
    else
	snprintf(Buff, sizeof(Buff), "%s ",
		 ShowEncodeName(SoftInfo->Name));

    if (SoftInfo->Version) {
	StrAppend(Buff, sizeof(Buff), " ");
	StrAppend(Buff, sizeof(Buff), SoftInfo->Version);
    }
    if (SoftInfo->Desc) {
	StrAppend(Buff, sizeof(Buff), " ");
	StrAppend(Buff, sizeof(Buff), SoftInfo->Desc);
    } else if (SoftInfo->DescVerbose) {
	StrAppend(Buff, sizeof(Buff), " ");
	/* Only print the first line */
	if (cp = strchr(SoftInfo->DescVerbose, '\n'))
	    snprintf(Buff2, sizeof(Buff2), " %.*s", 
		    cp-SoftInfo->DescVerbose, SoftInfo->DescVerbose);
	else
	    snprintf(Buff2, sizeof(Buff2), " %s", SoftInfo->DescVerbose);
	StrAppend(Buff, sizeof(Buff), Buff2);
    }

    ShowLine(OffSet, Buff);

    DetailsSoftInfo(SoftInfo, OffSet);
}

/*
 * Replace all actual newline chars with the string \n.
 * If anything is replaced in String, memory will be alloc'ed.
 * If nothing is replaced, String will be returned.
 */
static char *EscapeNewlines(String)
     char		       *String;
{
    int				Found = 0;
    register char	       *cp;
    char		       *bp;
    char		       *Buffer;
    size_t			Len;

    if (!String)
	return (char *) NULL;

    for (cp = String, Found = 0; cp && *cp; ++cp)
	if (*cp == '\n')
	    ++Found;

    if (!Found)
	/* 
	 * There's no need to scan String since we found nothing.
	 */
	return String;

    Len = cp - String + Found + 1;

    bp = Buffer = (char *) xcalloc(1, Len);

    for (cp = String; cp && *cp; ++cp) {
	if (*cp == '\n') {
	    *bp++ = '\\';
	    *bp++ = 'n';
	} else
	    *bp++ = *cp;
    }
    *bp = CNULL;

    return Buffer;
}

/*
 * Report format Display of a SoftInfo entry.
 */
static void DisplaySoftInfoReport(Soft, OffSet)
     SoftInfo_t		       *Soft;
     int			OffSet;
{
    static char		       *RptData[28];
    static char			SlaveBuff[BUFSIZ];
    Desc_t		       *Desc;
    SoftFileList_t	       *FileList;
    SoftFileData_t	       *File;
    SoftInfo_t		       *si;
    char		       *cp;

    if (!Soft || !Soft->Name)
	return;

    (void) memset(RptData, CNULL, sizeof(RptData));
    RptData[0] = PRTS(Soft->EntryType);
    /* 
     * This is a duplicate and shouldn't be here, but for backwards
     * compat it's left.
     */
    RptData[1] = PRTS(Soft->Name);
    RptData[2] = PRTS(Soft->Version);
    RptData[3] = PRTS(Soft->Revision);
    RptData[4] = PRTS(Soft->Desc);
    RptData[5] = PRTS(Soft->URL);
    RptData[6] = PRTS(Soft->License);
    RptData[7] = PRTS(Soft->Category);
    RptData[8] = PRTS(Soft->SubCategory);
    RptData[9] = PRTS(Soft->OSname);
    RptData[10] = PRTS(Soft->OSversion);
    RptData[11] = PRTS(Soft->Arch);
    RptData[12] = PRTS(Soft->ISArch);
    RptData[13] = PRTS(Soft->InstDate);
    RptData[14] = PRTS(Soft->BuildDate);
    RptData[15] = PRTS(Soft->ProdStamp);
    RptData[16] = PRTS(Soft->BaseDir);
    if (Soft->DiskUsage)
	RptData[17] = strdup(Ltoa(Soft->DiskUsage));
    RptData[18] = PRTS(Soft->VendorName);
    RptData[19] = PRTS(Soft->VendorEmail);
    RptData[20] = PRTS(Soft->VendorPhone);
    RptData[21] = PRTS(Soft->VendorStock);
    if (Soft->Master) {
	RptData[22] = PRTS(Soft->Master->Name);
	RptData[23] = PRTS(Soft->Master->Version);
	RptData[24] = PRTS(Soft->Master->Revision);
    }
    RptData[25] = PRTS(EscapeNewlines(Soft->DescVerbose));
    RptData[26] = PRTS(EscapeNewlines(Soft->Copyright));
    if (Soft->Next)
	RptData[27] = PRTS(Soft->Next->Name);
    if (Soft->Slaves) {
	SlaveBuff[0] = CNULL;
	for (si = Soft->Slaves; si; si = si->Next) {
	    if (SlaveBuff[0])
		strcat(SlaveBuff, ",");
	    if (strlen(si->Name) + strlen(SlaveBuff) + 3 >= sizeof(SlaveBuff))
		break;
	    strcat(SlaveBuff, si->Name);
	}
	if (SlaveBuff[0])
	    RptData[28] = SlaveBuff;
    }

    Report(CN_SOFTINFO, R_NAME, PRTS(Soft->Name), RptData,
	   sizeof(RptData) / sizeof(char *));

    /*
     * Report description info
     */
    (void) memset(RptData, CNULL, sizeof(RptData));
    for (Desc = Soft->DescList; Desc; Desc = Desc->Next) {
	RptData[0] = PRTS(Desc->Label);
	RptData[1] = PRTS(Desc->Desc);
	Report(CN_SOFTINFO, R_DESC, PRTS(Soft->Name), RptData, 2);
    }

    if (SwFiles) {
	/*
	 * Report all the file information
	 */
	(void) memset(RptData, CNULL, sizeof(RptData));
	for (FileList = Soft->FileList; FileList; 
	     FileList = FileList->Next) {
	    RptData[3] = RptData[5] = "";
	    if (File = FileList->FileData) {
		RptData[0] = SoftInfoFileType(File->Type);
		RptData[1] = PRTS(File->Path);
		RptData[2] = PRTS(File->LinkTo);
		if (File->FileSize && (cp = Ltoa(File->FileSize)))
		    RptData[3] = strdup(cp);
		RptData[4] = PRTS(File->MD5);
		RptData[5] = PRTS(File->CheckSum);
		if (File->PkgNames && 
		    (ArgvToStr(&cp, File->PkgNames, 0, ",", 0) == 0))
		    RptData[6] = cp;
		Report(CN_SOFTINFO, R_FILE, PRTS(Soft->Name), RptData, 6);
	    }
	}
    }
}

/*
 * Display a SoftInfo entry by calling the appropriate type of
 * format function.
 */
static void DisplaySoftInfo(SoftInfo, OffSet)
     SoftInfo_t		       *SoftInfo;
     int			OffSet;
{
    switch (FormatType) {
    case FT_TREE:
	DisplaySoftInfoTree(SoftInfo, OffSet);
	break;
    case FT_COLUMNS:
	DisplaySoftInfoColumns(SoftInfo, OffSet);
	break;
    case FT_REPORT:
	DisplaySoftInfoReport(SoftInfo, OffSet);
	break;
    }
}

/*
 * Display info about SoftInfo node and traverse tree.
 * --RECURSE--
 */
static void DisplaySoftInfoNode(SoftInfo, OffSet)
     SoftInfo_t		       *SoftInfo;
     int			OffSet;
{
    if (SoftInfo)
	DisplaySoftInfo(SoftInfo, OffSet);

    if (SoftInfo && SoftInfo->Slaves)
	DisplaySoftInfoNode(SoftInfo->Slaves, 
			(SoftInfo->Name) ? OffSet + OffSetAmt : 0);

    if (SoftInfo && SoftInfo->Next)
	DisplaySoftInfoNode(SoftInfo->Next, (SoftInfo->Name) ? OffSet : 0);
}
